import logging
import multiprocessing as mp
from pathlib import Path
import queue
import sys
from time import sleep
import traceback
from typing import Dict, List, Optional

from libmuscle.errors import ConfigurationError
from libmuscle.manager.instantiator import (
        CancelAllRequest, CrashedResult, create_instance_env, InstantiationRequest,
        Process, ProcessStatus, reconfigure_logging, ShutdownRequest)
from libmuscle.native_instantiator.agent_manager import AgentManager
from libmuscle.native_instantiator.global_resources import global_resources
from libmuscle.native_instantiator.run_script import make_script, prep_resources
from libmuscle.planner.resources import OnNodeResources, Resources
from ymmsl import MPICoresResReq, MPINodesResReq, ResourceRequirements, ThreadedResReq


_logger = logging.getLogger(__name__)


class NativeInstantiator(mp.Process):
    """Instantiates instances on the local machine."""
    def __init__(
            self, resources: mp.Queue, requests: mp.Queue, results: mp.Queue,
            log_records: mp.Queue, run_dir: Path) -> None:
        """Create a NativeInstantiator

        Args:
            resources: Queue for returning the available resources
            requests: Queue to take requests from
            results: Queue to communicate finished processes over
            log_messages: Queue to push log messages to
            run_dir: Run directory for the current run
        """
        super().__init__(name='NativeInstantiator')
        self._resources_out = resources
        self._requests_in = requests
        self._results_out = results
        self._log_records_out = log_records
        self._run_dir = run_dir

        self._processes: Dict[str, Process] = dict()

    def run(self) -> None:
        """Entry point for the process"""
        try:
            logs_dir = self._run_dir / 'logs'
            logs_dir.mkdir(exist_ok=True)

            self._agent_manager = AgentManager(logs_dir)

            reconfigure_logging(self._log_records_out)
            self._send_resources()
            self._main()

        except ConfigurationError as e:
            self._results_out.put(CrashedResult(e))

        except:     # noqa
            for line in traceback.format_exception(*sys.exc_info()):
                _logger.error(line)

            result = CrashedResult(sys.exc_info()[1])
            self._resources_out.put(result)
            self._results_out.put(result)

    def _main(self) -> None:
        """Main function for the background process.

        This accepts requests for instantiating jobs, stopping them, or shutting down.
        Results of finished jobs are returned via the results queue.
        """
        shutting_down = False
        done = False
        while not done:
            while not shutting_down:
                try:
                    request = self._requests_in.get_nowait()
                    if isinstance(request, ShutdownRequest):
                        _logger.debug('Got ShutdownRequest')
                        shutting_down = True

                    elif isinstance(request, CancelAllRequest):
                        _logger.debug('Got CancelAllRequest')
                        self._agent_manager.cancel_all()

                    elif isinstance(request, InstantiationRequest):
                        _logger.debug('Got InstantiationRequest')
                        if not shutting_down:
                            self._instantiate(request)

                except queue.Empty:
                    break

            self._report_failed_processes()
            self._report_finished_processes()

            if shutting_down:
                _logger.debug(f'Remaining processes: {self._processes}')
                done = not self._processes

            if not done:
                sleep(0.1)

        self._agent_manager.shutdown()

    def _send_resources(self) -> None:
        """Detect resources and report them to the manager.

        We have potentially two sources of truth here: the Slurm environment variables
        and what the agents report based on what they're bound to. These should be
        consistent, but we check that and then try to be conservative to try to not
        step outside our bounds even if the cluster doesn't constrain processes to their
        assigned processors.
        """
        already_logged_smt = False
        resources = Resources()

        agent_res = self._agent_manager.get_resources()

        env_ncpus = dict(
                zip(global_resources().nodes, global_resources().logical_cpus_per_node)
                )

        for node_name in env_ncpus:
            if node_name not in agent_res.nodes():
                _logger.warning(
                        f'The environment suggests we should have node {node_name},'
                        ' but no agent reported running on it. We won''t be able'
                        ' to use this node.')
            else:
                env_nncpus = env_ncpus[node_name]
                ag_nncores = len(agent_res[node_name].cpu_cores)
                ag_nnthreads = len(list(agent_res[node_name].hwthreads()))

                if ag_nncores != ag_nnthreads and ag_nnthreads == env_nncpus:
                    if not already_logged_smt:
                        _logger.info(
                                'Detected SMT (hyperthreading) as available and'
                                ' enabled. Note that MUSCLE3 will assign whole cores to'
                                ' each thread or MPI process.')
                        already_logged_smt = True

                    resources.add_node(agent_res[node_name])

                elif ag_nncores < env_nncpus:
                    _logger.warning(
                            f'Node {node_name} should have {env_nncpus} cores'
                            f' available, but the agent reports only {ag_nncores}'
                            f' available to it. We\'ll use the {ag_nncores} we seem to'
                            ' have.')

                    resources.add_node(agent_res[node_name])

                elif env_nncpus < ag_nncores:
                    _logger.warning(
                            f'Node {node_name} should have {env_nncpus} cores'
                            f' available, but the agent reports {ag_nncores} available'
                            ' to it. Maybe the cluster does not constrain resources?'
                            f' We\'ll use the {env_nncpus} that we should have got.')
                    resources.add_node(
                            OnNodeResources(
                                node_name,
                                agent_res[node_name].cpu_cores.get_first_cores(
                                    env_nncpus)))

                else:
                    # no SMT, agent matches environment
                    resources.add_node(agent_res[node_name])

        for node in agent_res:
            if node.node_name not in env_ncpus:
                _logger.warning(
                        f'An agent is running on node {node.node_name} but the'
                        ' environment does not list it as ours. It seems that the'
                        ' node\'s hostname does not match what SLURM calls it. We will'
                        ' not use this node, because we\'re not sure it\'s really ours'
                        ' or if it is, how many resources on it we can use.'
                        )

        self._resources_out.put(resources)

    def _instantiate(self, request: InstantiationRequest) -> None:
        """Instantiate an implementation according to the request."""
        name = str(request.instance)

        env = create_instance_env(
                request.instance, request.implementation.base_env,
                request.implementation.env)
        self._add_resources(env, request.res_req)

        rankfile = request.instance_dir / 'rankfile'

        if global_resources().on_cluster():
            rankfile_contents, resource_env = prep_resources(
                  request.implementation.execution_model, request.resources,
                  rankfile)

            if rankfile_contents:
                with rankfile.open('w') as f:
                    f.write(rankfile_contents)
                env['MUSCLE_RANKFILE'] = str(rankfile)

            env.update(resource_env)

        run_script_file = self._write_run_script(request, rankfile)
        args = [str(run_script_file)]

        self._processes[name] = Process(request.instance, request.resources)

        _logger.debug(f'Instantiating {name} on {request.resources}')
        try:
            self._agent_manager.start(
                    request.resources.by_rank[0].node_name,
                    name, request.work_dir, args, env,
                    request.stdout_path, request.stderr_path)
            self._processes[name].status = ProcessStatus.RUNNING

        except Exception as e:
            _logger.warning(f'Instance {name} failed to start: {e}')
            self._processes[name].status = ProcessStatus.ERROR
            self._processes[name].error_msg = f'Instance failed to start: {e}'

    def _write_run_script(
            self, request: InstantiationRequest, rankfile: Optional[Path]) -> Path:
        """Create and write out the run script and return its location."""
        # TODO: Only write out once for each implementation
        if request.implementation.script:
            run_script = request.implementation.script
        else:
            run_script = make_script(
                    request.implementation, request.res_req,
                    request.work_dir, not global_resources().on_cluster(), rankfile)

        run_script_file = request.instance_dir / 'run_script.sh'

        with run_script_file.open('w') as f:
            f.write(run_script)

        run_script_file.chmod(0o700)
        return run_script_file

    def _add_resources(
            self, env: Dict[str, str], res_req: ResourceRequirements) -> None:
        """Add resource env vars to the given env."""
        if isinstance(res_req, ThreadedResReq):
            num_threads = res_req.threads
        elif isinstance(res_req, (MPICoresResReq, MPINodesResReq)):
            num_threads = res_req.threads_per_mpi_process

        env['MUSCLE_THREADS'] = str(num_threads)
        env['OMP_NUM_THREADS'] = str(num_threads)

        num_mpi_processes: Optional[int] = None
        if isinstance(res_req, MPICoresResReq):
            num_mpi_processes = res_req.mpi_processes
        elif isinstance(res_req, MPINodesResReq):
            num_mpi_processes = res_req.nodes * res_req.mpi_processes_per_node

        if num_mpi_processes is not None:
            env['MUSCLE_MPI_PROCESSES'] = str(num_mpi_processes)

    def _report_failed_processes(self) -> None:
        """Get processes that failed to start and report their status."""
        failed_processes: List[str] = list()

        for name, process in self._processes.items():
            if process.status == ProcessStatus.ERROR:
                self._results_out.put(process)
                failed_processes.append(name)

        for name in failed_processes:
            del self._processes[name]

    def _report_finished_processes(self) -> None:
        """Get finished processes and report back their status."""
        for name, exit_code in self._agent_manager.get_finished():
            process = self._processes[name]
            if process.status == ProcessStatus.RUNNING:
                if exit_code == 0:
                    process.status = ProcessStatus.SUCCESS
                else:
                    process.status = ProcessStatus.ERROR
                    process.error_msg = 'Instance returned a non-zero exit code'
            process.exit_code = exit_code
            self._results_out.put(process)
            del self._processes[name]



from typing import List, Optional, Set

from ymmsl import SettingValue, Reference, Settings


def has_setting_type(value: SettingValue, typ: str) -> bool:
    """Checks whether the value has the given type.

    Args:
        value: A setting value.
        typ: A setting type. Valid values are 'str', 'int', 'float',
                'bool', '[int]', '[float]', and '[[float]]'.

    Returns:
        True if the type of value matches typ.

    Raises:
        ValueError: If the type specified is not valid.
    """
    par_type_to_type = {
            'str': str,
            'int': int,
            'float': float,
            'bool': bool
            }

    if typ in par_type_to_type:
        return isinstance(value, par_type_to_type[typ])
    elif typ == '[int]':
        if isinstance(value, list):
            if len(value) == 0 or isinstance(value[0], int):
                # We don't check everything here, the yMMSL loader does
                # a full type check, so we just need to discriminate.
                return True
    elif typ == '[float]':
        if isinstance(value, list):
            if len(value) == 0 or isinstance(value[0], float):
                # We don't check everything here, the yMMSL loader does
                # a full type check, so we just need to discriminate.
                return True
        return False
    elif typ == '[[float]]':
        if isinstance(value, list):
            if len(value) == 0 or isinstance(value[0], list):
                # We don't check everything here, the yMMSL loader does
                # a full type check, so we just need to discriminate.
                return True
        return False
    raise ValueError('Invalid setting type specified: {}'.format(typ))


class SettingsManager:
    """Manages the current settings for a component instance.
    """
    def __init__(self) -> None:
        """Create a SettingsManager.

        Initialises the base and overlay layers to an empty
        Settings object.

        A SettingsManager has two layers of settings, a base
        layer that contains an immutable collection of settings set in
        the simulation's yMMSL description, and an overlay layer that
        holds settings that have been set at run-time.

        Attributes:
            base: The base layer.
            overlay: The overlay layer.
        """
        self.base = Settings()
        self.overlay = Settings()

    def list_settings(self, instance_id: Reference) -> List[str]:
        """Returns the names of all the settings.

        This returns the names of all the settings, as the model would
        pass them to request settings. It returns

        - <setting_name> as-is
        - <instance_id>.<setting_name> as <setting_name>
        - <other_id>.<setting_name> not at all
        - <setting>.<setting> not at all

        Note that we don't return global settings with multipart names.
        Those are legal, but currently indistinguishable from settings
        intended for other components. We're not actually telling
        anyone that they're legal, so we can probably get away with
        that. If it becomes an issue, we'll have to get a list of
        instance ids from the manager so we can recognise them
        correctly.

        Args:
            instance_id: Our instance id.

        Return:
            A list of setting names.
        """
        def extract_names(settings: Settings) -> Set[str]:
            result: Set[str] = set()
            for name in settings:
                if len(name) == 1:
                    result.add(str(name))
                else:
                    id_len = len(instance_id)
                    if len(name) > id_len:
                        if name[:id_len] == instance_id:
                            result.add(str(name[id_len:]))
            return result

        names = extract_names(self.base)
        names.update(extract_names(self.overlay))
        return sorted(names)

    def get_setting(self, instance: Reference, setting_name: Reference,
                    typ: Optional[str] = None) -> SettingValue:
        """Returns the value of a setting.

        Args:
            instance: The instance that this value is for.
            setting_name: The name of the setting to get the value of.
            typ: An optional type designation; if specified the type
                    is checked for a match before returning. Valid
                    values are 'str', 'int', 'float', 'bool',
                    '[float]' and '[[float]]'.

        Raises:
            KeyError: If the setting has not been set.
            TypeError: If the setting was set to a value that does
                    not match `typ`.
            ValueError: If an invalid value was specified for `typ`
        """
        for i in range(len(instance), -1, -1):
            if i > 0:
                name = instance[:i] + setting_name
            else:
                name = setting_name

            if name in self.overlay:
                value = self.overlay[name]
                break
            elif name in self.base:
                value = self.base[name]
                break
        else:
            raise KeyError(('Value for setting "{}" was not set.'.format(
                setting_name)))

        if typ is not None:
            if not has_setting_type(value, typ):
                raise TypeError('Value for setting "{}" is of type {},'
                                ' where {} was expected.'.format(
                                    name, type(value), typ))
        return value



from libmuscle.communicator import Message
from libmuscle.grid import Grid
from libmuscle.instance import Instance, InstanceFlags
from libmuscle.manager.profile_database import ProfileDatabase
from libmuscle.version import __version__
from libmuscle import runner


# Note that libmuscle.version above is created by the build system; it's okay
# that it's not present.

__all__ = [
        '__version__', 'Grid', 'Instance', 'InstanceFlags', 'Message',
        'ProfileDatabase', 'runner']


# export InstanceFlag members to the module namespace
# adapted from https://github.com/python/cpython/blob/3.10/Lib/re.py#L179
globals().update(InstanceFlags.__members__)
__all__.extend(InstanceFlags.__members__)



from datetime import datetime
from pathlib import Path
import sys
import textwrap
import traceback
from typing import Optional, Sequence

import click
from yaml.scanner import ScannerError
from yatiml import RecognitionError
import ymmsl
from ymmsl import Identifier, PartialConfiguration

from libmuscle.manager.logger import last_lines
from libmuscle.manager.manager import Manager
from libmuscle.manager.run_dir import RunDir


@click.command(no_args_is_help=True)
@click.argument(
        'ymmsl_files', nargs=-1, required=True, type=click.Path(
            exists=True, file_okay=True, dir_okay=False, readable=True,
            allow_dash=True, resolve_path=True))
@click.option(
        '--log-level', nargs=1, type=str, default='INFO', show_default=True,
        help='Set the local log level. Try DEBUG for (much) more output.')
@click.option(
        '--run-dir', nargs=1, type=click.Path(
            exists=True, file_okay=False, dir_okay=True, readable=True,
            writable=True, allow_dash=False, resolve_path=True),
        help=(
            'Directory to write logs, run metadata and output to. By default,'
            ' a directory with the name run_<model>_<date>-<time> will be'
            ' created.')
        )
@click.option(
        '--location-file', nargs=1, type=click.Path(
            exists=False, file_okay=True, dir_okay=True, readable=False,
            allow_dash=True),
        help=(
            'File to write the network location of the manager to if'
            ' --start-all is not specified. If a relative path is given,'
            ' then it will be resolved relative to the directory in which'
            ' the manager was started.'
            '\b\n\n'
            ' The manager will write to this file a single line of text,'
            ' which should be passed to the instances via the'
            ' MUSCLE_MANAGER environment variable or on their command line'
            ' using the --muscle-manager=<contents> option.'
            '\b\n\n'
            ' If --start-all is not specified and --location-file is also'
            ' not given, then the location will be printed on standard'
            ' output.')
        )
@click.option(
        '--start-all/--no-start-all', default=False, help=(
            'Start all submodel instances listed in the configuration file(s).'
            )
        )
def manage_simulation(
        ymmsl_files: Sequence[str],
        start_all: bool,
        run_dir: Optional[str],
        log_level: Optional[str],
        location_file: Optional[str]
        ) -> None:
    """Run the MUSCLE3 Manager.

    The MUSCLE manager manages a coupled simulation. It can start the
    various submodels, and it helps them to find and connect to each
    other to exchange messages. The manager also distributes settings
    from the configuration file(s) to the submodel instances.

    The muscle_manager command takes a list of configuration files in
    yMMSL format (see the online documentation). The files are combined
    in the order given, so if for example a setting is given multiple
    times, then the value in the last file in the list to mention it is
    used.
    """
    configuration = PartialConfiguration()
    for path in ymmsl_files:
        configuration.update(load_configuration(path))

    try:
        if start_all:  # Do a full consistency check
            configuration.as_configuration().check_consistent()
        else:  # Only require that a Model exists and is consistent
            if configuration.model is None:
                raise ValueError('Model section is missing from the configuration.')
            if not isinstance(configuration.model, ymmsl.Model):
                raise ValueError('Model section from the configuration is incomplete.')
            configuration.model.check_consistent()
    except Exception as exc:
        print('Failed to start the simulation, found a configuration error:\n' +
              textwrap.indent(str(exc), 4*' '), file=sys.stderr)
        sys.exit(1)

    if run_dir is None:
        if configuration.model is not None:
            model_name = configuration.model.name
        else:
            model_name = Identifier('model')
        timestamp = datetime.now()
        timestr = timestamp.strftime('%Y%m%d_%H%M%S')
        run_dir_path = Path.cwd() / 'run_{}_{}'.format(model_name, timestr)
    else:
        run_dir_path = Path(run_dir).resolve()

    run_dir_obj = RunDir(run_dir_path)
    if start_all:
        manager = Manager(configuration, run_dir_obj, log_level)
        try:
            manager.start_instances()
        except Exception as exc:
            manager.stop()
            print('Failed to start the simulation:', file=sys.stderr)
            print(textwrap.indent(str(exc), 4*' '), file=sys.stderr)
            print(file=sys.stderr)
            print('Check the manager log for more details:', file=sys.stderr)
            print('   ', run_dir_path / 'muscle3_manager.log', file=sys.stderr)
            sys.exit(1)

    else:
        if run_dir is None:
            manager = Manager(configuration, None, log_level)
        else:
            manager = Manager(configuration, run_dir_obj, log_level)

        server_location = manager.get_server_location()
        if location_file is None:
            print(server_location, flush=True)
        else:
            Path(location_file).write_text(server_location)

    success = manager.wait()

    if not success:
        log_file = run_dir_path / 'muscle3_manager.log'

        print()
        print('An error occurred during execution, and the simulation was')
        print('shut down. The manager log should tell you what happened.')
        print('Here are the final lines of the manager log:')
        print()
        print('-' * 80)
        print(last_lines(log_file, 50), '    ')
        print('-' * 80)
        print()
        print('You can find the full log at')
        print(str(log_file))
        print()
    else:
        print('Simulation completed successfully.')
        try:
            rel_run_dir = run_dir_path.relative_to(Path.cwd())
            print(f'Output may be found in {rel_run_dir}')
        except ValueError:
            print(f'Output may be found in {run_dir_path}')

    sys.exit(0 if success else 1)


def load_configuration(path: str) -> PartialConfiguration:
    """Load and parse a configuration file.

    Annotates error messages for easier debugging by users.
    """
    with open(path, 'r') as f:
        try:
            return ymmsl.load(f)
        except ScannerError as exc:  # capture yaml errors
            # PyYAML error messages are not very user friendly, but there is not
            # too much we can do about it
            print(f"Syntax error while loading configuration file '{path}':",
                  file=sys.stderr)
            print(textwrap.indent(str(exc), 4*' '), file=sys.stderr)
            sys.exit(1)
        except RecognitionError as exc:
            # capture yatiml errors:
            # - ymmsl syntax errors, like mapping instead of lists, misspelled keys, ...
            # - value errors thrown by constructors (e.g. specifying duplicate port
            #   names)
            print(f"Recognition error while loading configuration file '{path}':",
                  file=sys.stderr)
            print(textwrap.indent(str(exc), 4*' '), file=sys.stderr)
            sys.exit(1)
        except Exception:
            # Any other error is not anticipated
            print(f"Error while loading configuration file '{path}':",
                  file=sys.stderr)
            traceback.print_exc()
            print(file=sys.stderr)
            print('This error could indicate a bug in libmuscle,'
                  ' please make an issue on GitHub.', file=sys.stderr)
            sys.exit(1)


if __name__ == '__main__':
    manage_simulation()



import logging
import os
import psutil
from socket import gethostname
import sys
from time import sleep
from typing import Dict, Set, Tuple

from libmuscle.native_instantiator.process_manager import ProcessManager
from libmuscle.native_instantiator.agent.map_client import MAPClient
from libmuscle.native_instantiator.agent.agent_commands import (
        CancelAllCommand, ShutdownCommand, StartCommand)
from libmuscle.planner.resources import Core, CoreSet, OnNodeResources


_logger = logging.getLogger(__name__)


class Agent:
    """Runs on a compute node and starts processes there."""
    def __init__(self, node_name: str, server_location: str) -> None:
        """Create an Agent.

        Args:
            node_name: Name (hostname) of this node
            server_location: MAP server of the manager to connect to
        """
        _logger.info(f'Agent at {node_name} starting')

        self._process_manager = ProcessManager()

        self._node_name = node_name

        _logger.info(f'Connecting to manager at {server_location}')
        self._server = MAPClient(self._node_name, server_location)
        _logger.info('Connected to manager')

    def run(self) -> None:
        """Execute commands and monitor processes."""
        _logger.info('Reporting resources')
        self._server.report_resources(self._inspect_resources())

        shutting_down = False
        while not shutting_down:
            command = self._server.get_command()
            if isinstance(command, StartCommand):
                _logger.info(f'Starting process {command.name}')
                _logger.debug(f'Args: {command.args}')
                _logger.debug(f'Env: {command.env}')

                self._process_manager.start(
                        command.name, command.work_dir, command.args, command.env,
                        command.stdout, command.stderr)
            elif isinstance(command, CancelAllCommand):
                _logger.info('Cancelling all instances')
                self._process_manager.cancel_all()

            elif isinstance(command, ShutdownCommand):
                # check that nothing is running
                shutting_down = True
                _logger.info('Agent shutting down')

            finished = self._process_manager.get_finished()
            if finished:
                for name, exit_code in finished:
                    _logger.info(f'Process {name} finished with exit code {exit_code}')
                self._server.report_result(finished)

            sleep(0.1)

    def _inspect_resources(self) -> OnNodeResources:
        """Inspect the node to find resources and report on them.

        The terminology for identifying processors gets very convoluted, with Linux,
        Slurm, OpenMPI and IntelMPI all using different terms, or sometimes the same
        terms for different things. See the comment in planner/resources.py for what is
        what and how we use it.

        Returns:
            A dict mapping resource types to resource descriptions.
        """
        if hasattr(os, 'sched_getaffinity'):
            hwthreads_by_core_tuple: Dict[Tuple[int, int], Set[int]] = dict()

            # these are the logical hwthread ids that we can use
            hwthread_ids = list(os.sched_getaffinity(0))

            for hwthread_id in hwthread_ids:
                topdir = f'/sys/devices/system/cpu/cpu{hwthread_id}/topology'
                with open(f'{topdir}/core_id', 'r') as f:
                    # this gets the logical core id for the hwthread
                    core_id = int(f.read())
                with open(f'{topdir}/physical_package_id', 'r') as f:
                    # this gets the die/socket id for the hwthread
                    package_id = int(f.read())

                core_tuple = (package_id, core_id)
                hwthreads_by_core_tuple.setdefault(core_tuple, set()).add(hwthread_id)

            core_lgid_hwthreads = [
                    (i, hwthreads_by_core_tuple[core_tuple])
                    for i, core_tuple in enumerate(
                        sorted(hwthreads_by_core_tuple.keys()))]

            cores = CoreSet((
                    Core(core_lgid, hwthreads)
                    for core_lgid, hwthreads in core_lgid_hwthreads))

        else:
            # MacOS doesn't support thread affinity, but older Macs with Intel
            # processors do have SMT. Getting the hwthread to core mapping is not so
            # easy, and if we're running on macOS then we're not on a cluster and don't
            # do binding anyway. So we're going to get the number of hwthreads and the
            # number of cores here, and synthesise a mapping that may be wrong, but will
            # at least represent the number of cores and threads per core correctly.
            nhwthreads = psutil.cpu_count(logical=True)
            ncores = psutil.cpu_count(logical=False)

            if nhwthreads is None:
                if ncores is not None:
                    _logger.warning(
                            'Could not determine number of hwthreads, assuming no SMT')
                    nhwthreads = ncores
                else:
                    _logger.warning(
                            'Could not determine CPU configuration, assuming a single'
                            ' core')
                    ncores = 1
                    nhwthreads = 1
            elif ncores is None:
                _logger.warning(
                        'Could not determine number of cores, assuming no SMT')
                ncores = nhwthreads

            hwthreads_per_core = nhwthreads // ncores

            if ncores * hwthreads_per_core != nhwthreads:
                # As far as I know, there are no Macs with heterogeneous SMT, like in
                # the latest Intel CPUs.
                _logger.warning(
                        'Only some cores seem to have SMT, core ids are probably'
                        ' wrong. If this is a cluster then this will cause problems,'
                        ' please report an issue on GitHub and report the machine and'
                        ' what kind of OS and hardware it has. If we\'re running on a'
                        ' local machine, then this won\'t affect the run, but I\'d'
                        ' still appreciate an issue, because it is unexpected for sure.'
                        )

            cores = CoreSet((
                    Core(
                        cid,
                        set(range(
                            cid * hwthreads_per_core, (cid + 1) * hwthreads_per_core))
                        )
                    for cid in range(ncores)
                    ))

        resources = OnNodeResources(self._node_name, cores)
        _logger.info(f'Found resources: {resources}')
        return resources


def configure_logging(node_name: str, log_level: int) -> None:
    """Make us output logs to a custom log file."""
    fmt = '%(asctime)s %(levelname)s %(message)s'
    formatter = logging.Formatter(fmt)

    handler = logging.FileHandler(f'muscle3_agent_{node_name}.log', mode='w')
    handler.setFormatter(formatter)

    # Find and remove default handler to disable automatic console output
    # Testing for 'stderr' in the stringified version is not nice, but
    # seems reliable, and doesn't mess up pytest's caplog mechanism while
    # it also doesn't introduce a runtime dependency on pytest.
    logging.getLogger().handlers = [
            h for h in logging.getLogger().handlers
            if 'stderr' not in str(h)]

    logging.getLogger().addHandler(handler)

    logging.getLogger().setLevel(log_level)


if __name__ == '__main__':
    node_name = gethostname()
    server_location = sys.argv[1]
    log_level = int(sys.argv[2])

    configure_logging(node_name, log_level)

    agent = Agent(node_name, server_location)
    agent.run()



import logging
from pathlib import Path
from typing import List, Optional

from libmuscle.logging import LogLevel, Timestamp
from libmuscle.util import extract_log_file_location


class Formatter(logging.Formatter):
    """A custom formatter that can format remote messages."""
    def usesTime(self) -> bool:
        """Tells the formatter to make asctime available."""
        return True

    def formatMessage(self, record: logging.LogRecord) -> str:
        """Formats a message for a record.

        If the record contains a time_stamp attribute, assumes that it
        is a remote record and formats accordingly, otherwise formats
        as a local record.

        Args:
            record: The LogRecord to format.

        Returns:
            The formatted message.
        """
        if 'instance' in record.__dict__:
            return (
                    '%(instance)-14s %(iasctime)-15s %(levelname)-7s'
                    ' %(name)s: %(message)s' % record.__dict__)
        return (
                'muscle_manager %(asctime)s %(levelname)-7s %(name)s:'
                ' %(message)s' % record.__dict__)


class Logger:
    """The MUSCLE3 Manager Logger component.

    The Logger component configures the local logging system to output
    to the central log file, and it accepts messages from remote
    instances to write to it as well. Log levels are also set here.

    """
    def __init__(
            self, log_dir: Optional[Path] = None,
            log_level: Optional[str] = None) -> None:
        """Create a Logger.

        Log levels may be any of the Python predefined log levels, i.e.
        critical, error, warning, info and debug, and they're
        case_insensitive.

        Args:
            log_dir: Directory to write the log file into.
            log_level: Log level to set.
        """

        if log_dir is None:
            log_dir = Path.cwd()
        logfile = extract_log_file_location('muscle3_manager.log')
        if logfile is None:
            logfile = log_dir / 'muscle3_manager.log'
        self._local_handler = logging.FileHandler(str(logfile), mode='w')
        self._local_handler.setFormatter(Formatter())

        # Find and remove default handler to disable automatic console output
        # Testing for 'stderr' in the stringified version is not nice, but
        # seems reliable, and doesn't mess up pytest's caplog mechanism while
        # it also doesn't introduce a runtime dependency on pytest.
        logging.getLogger().handlers = [
                h for h in logging.getLogger().handlers
                if 'stderr' not in str(h)]

        # add our own
        logging.getLogger().addHandler(self._local_handler)

        # set the root logger to log everything, so remote messages
        # are always logged
        logging.getLogger().setLevel(logging.NOTSET)

        # set Manager log level
        if log_level is None:
            log_level = 'INFO'
        else:
            log_level = log_level.upper()

        logging.getLogger('libmuscle').setLevel(log_level)

        # YAtiML should be pretty reliable, and if there is an issue
        # then we can easily load the problem file in Python by hand
        # using the yMMSL library and set the log level there.
        logging.getLogger('yatiml').setLevel(logging.WARNING)

    def close(self) -> None:
        logging.getLogger().removeHandler(self._local_handler)

    def log_message(
            self,
            instance_id: str,
            timestamp: Timestamp,
            level: LogLevel,
            text: str) -> None:
        """Log a message.

        Args:
            instance_id: Identifier of the instance that generated the \
                    message.
            timestamp: Time when this log message was generated, \
                    according to the clock on that machine.
            level: The log level of the message.
            text: The message text.
        """
        logger = logging.getLogger(instance_id)
        logger.log(
                level.as_python_level(),
                text,
                extra={
                    'instance': instance_id,
                    'iasctime': timestamp.to_asctime()})


def last_lines(file: Path, count: int) -> str:
    """Utility function that returns the last lines of a text file.

    This opens the file and returns the final `count` lines. It reads
    at most 10000 bytes, to avoid memory problems if the file contains
    e.g. a large amount of binary data.

    Args:
        file: The file to read
        count: Number of lines to read

    Return:
        A string of at most 10000 bytes, containing at most `count`
        newlines.
    """
    if not file.exists():
        return ''

    file_size = file.stat().st_size
    start_point = max(file_size - 10000, 0)

    lines: List[str] = []
    with file.open('r') as f:
        f.seek(start_point)
        f.readline()    # skip partial line
        line = f.readline()
        while line:
            lines.append(line)
            line = f.readline()

    return '\n' + ''.join(lines[-count:])



from unittest.mock import MagicMock

import pytest
from ymmsl import Reference

from libmuscle.mcp.tcp_transport_server import TcpTransportServer
from libmuscle.mpp_message import MPPMessage
from libmuscle.outbox import Outbox
from libmuscle.post_office import PostOffice


@pytest.fixture
def receiver():
    return Reference('test_receiver.test_port')


@pytest.fixture
def post_office(receiver):
    class MockPO(PostOffice):
        outboxes = {receiver: Outbox()}

        def get_message(self, receiver: Reference) -> MPPMessage:
            return self.outboxes[receiver].retrieve()

    return MockPO()


@pytest.fixture
def tcp_transport_server():
    server = TcpTransportServer(MagicMock())
    yield server
    server.close()



import logging
import os
import socket

from libmuscle import Instance, Message
from ymmsl import Operator


def log_location() -> None:
    """Log where we are running so that the test can check for it."""
    print(socket.gethostname())
    print(','.join(map(str, sorted(os.sched_getaffinity(0)))))


def component() -> None:
    """A simple dummy component.

    This sends and receives on all operators, allowing different coupling patterns
    with a single program.
    """
    instance = Instance({
            Operator.F_INIT: ['init_in'],
            Operator.O_I: ['inter_out'],
            Operator.S: ['inter_in'],
            Operator.O_F: ['final_out']})

    while instance.reuse_instance():
        # F_INIT
        steps = instance.get_setting('steps', 'int')

        instance.receive('init_in', default=Message(0.0))

        for step in range(steps):
            # O_I
            instance.send('inter_out', Message(step))

            # S
            instance.receive('inter_in', default=Message(0.0))

        # O_F
        instance.send('final_out', Message(steps))


if __name__ == '__main__':
    logging.basicConfig()
    logging.getLogger().setLevel(logging.INFO)

    log_location()
    component()



import logging
from pathlib import Path
from typing import Dict, Iterable, List, Optional, Tuple

from libmuscle.errors import ConfigurationError
from libmuscle.native_instantiator.slurm import slurm
from libmuscle.planner.planner import ResourceAssignment
from ymmsl import (
        BaseEnv, ExecutionModel, Implementation, MPICoresResReq, MPINodesResReq,
        ResourceRequirements, ThreadedResReq)


def direct_prep_resources(resources: ResourceAssignment) -> Tuple[str, Dict[str, str]]:
    """Create resources for a non-MPI program with taskset.

    Taskset expects a set of hwthreads on the command line, either as a comma-separated
    list or as a hexadecimal mask. We generate both here and set two environment
    variables.

    Args:
        resources: The resource assignment to describe

    Return:
        No rank file, and a set of environment variables.
    """
    env: Dict[str, str] = dict()
    only_node_hwthreads_list = list(resources.by_rank[0].hwthreads())

    env['MUSCLE_BIND_LIST'] = ','.join(map(str, only_node_hwthreads_list))

    mask_int = sum((1 << c for c in only_node_hwthreads_list))
    env['MUSCLE_BIND_MASK'] = format(mask_int, 'X')

    return '', env


def openmpi_prep_resources(resources: ResourceAssignment) -> Tuple[str, Dict[str, str]]:
    """Create resource description for OpenMPI mpirun

    Args:
        resources: The resource assignment to describe

    Return:
        The contents of the rankfile, and a set of environment variables
    """
    ranklines: List[str] = list()
    all_cores = (
            (node_res, ','.join(map(str, sorted(node_res.hwthreads()))))
            for node_res in resources.by_rank)

    for i, (node_res, hwthreads) in enumerate(all_cores):
        ranklines.append(f'rank {i}={node_res.node_name} slot={hwthreads}')

    rankfile = '\n'.join(ranklines) + '\n'

    return rankfile, dict()


def impi_prep_resources(resources: ResourceAssignment) -> Tuple[str, Dict[str, str]]:
    """Create resource description for Intel MPI mpirun

    Args:
        resources: The resource assignment to describe

    Return:
        The contents of the machinefile, and a set of environment variables
    """
    env: Dict[str, str] = dict()
    machine_nodes: List[str] = list()
    pin_masks: List[int] = list()

    for rank, res in enumerate(resources.by_rank):
        machine_nodes.append(res.node_name)
        pin_masks.append(sum((1 << c for c in res.hwthreads())))

    # coalesce machine lines
    proc_counts = [1] * len(machine_nodes)
    i = 1
    while i < len(machine_nodes):
        if machine_nodes[i-1] == machine_nodes[i]:
            del machine_nodes[i]
            proc_counts[i-1] += proc_counts[i]
            del proc_counts[i]
        else:
            i += 1

    machinefile = '\n'.join(
            (f'{m}:{c}' for m, c in zip(machine_nodes, proc_counts))) + '\n'

    # disable pinning to SLURM-specified resources
    # env['I_MPI_PIN_RESPECT_CPUSET'] = '0'
    env['I_MPI_JOB_RESPECT_PROCESS_PLACEMENT'] = 'off'

    # which cores to bind each rank to
    pin_masks_str = ','.join(format(mask, '#x') for mask in pin_masks)
    env['I_MPI_PIN_DOMAIN'] = f'[{pin_masks_str}]'

    # I_MPI_PIN_DOMAIN=[55,aa]
    # pins the first rank to 0,2,16,18 and the second to 1,3,17,19
    # I_MPI_PIN_PROCESSOR_LIST=0,1,5,6
    # pins rank 0 to core 0, rank 1 to core 1, rank 2 to core 5, rank 3 to core 6
    # machinefile:
    # host1:2
    # host2:4
    # runs two processes on host1 and four on host2
    return machinefile, env


def mpich_prep_resources(resources: ResourceAssignment) -> Tuple[str, Dict[str, str]]:
    """Create resource description for MPICH mpirun

    Args:
        resources: The resource assignment to describe

    Return:
        The contents of the machinefile, and a set of environment variables
    """
    # No env vars, but rankfile
    raise NotImplementedError()


def srun_prep_resources(
        resources: ResourceAssignment, rankfile_location: Path
        ) -> Tuple[str, Dict[str, str]]:
    """Create resource description for srun

    Args:
        resources: The resources to describe
        rankfile_location: Location where the rankfile will be written

    Return:
        The contents of the hostfile, and a set of environment variables
    """
    hostfile = '\n'.join((
        node_res.node_name for node_res in resources.by_rank
        for _ in node_res.hwthreads()))

    env = {'SLURM_HOSTFILE': str(rankfile_location)}

    def core_mask(hwthreads: Iterable[int]) -> str:
        mask = sum((1 << hwthread) for hwthread in hwthreads)
        return format(mask, '#x')

    bind_str = ','.join([
        core_mask(node_res.hwthreads()) for node_res in resources.by_rank])

    env['SLURM_CPU_BIND'] = f'verbose,mask_cpu:{bind_str}'

    return hostfile, env


def prep_resources(
        model: ExecutionModel, resources: ResourceAssignment, rankfile_location: Path
        ) -> Tuple[str, Dict[str, str]]:
    """Create resource description for the given execution model.

    Args:
        model: The execution model to generate a description for
        resources: The resource assignment to describe
        rankfile_location: Path to where the rankfile will be written

    Return:
        The contents of the rank/machine/hostfile, and a set of environment variables.
    """
    if model == ExecutionModel.DIRECT:
        return direct_prep_resources(resources)
    elif model == ExecutionModel.OPENMPI:
        return openmpi_prep_resources(resources)
    elif model == ExecutionModel.INTELMPI:
        return impi_prep_resources(resources)
    elif model == ExecutionModel.SRUNMPI:
        return srun_prep_resources(resources, rankfile_location)
    # elif model == ExecutionModel.MPICH:
    #     return mpich_prep_resources(resources)
    raise RuntimeError(
            f'Impossible execution model {model}, please create an issue on GitHub')


def num_mpi_tasks(res_req: ResourceRequirements) -> int:
    """Determine the number of MPI tasks to be started.

    For non-MPI resource requirements, returns 1.

    Args:
        res_req: Resource requirements to analyse.
    """
    if isinstance(res_req, ThreadedResReq):
        return 1
    elif isinstance(res_req, MPICoresResReq):
        return res_req.mpi_processes
    elif isinstance(res_req, MPINodesResReq):
        return res_req.nodes * res_req.mpi_processes_per_node
    raise RuntimeError('Invalid ResourceRequirements')


def local_command(implementation: Implementation, enable_debug: bool) -> str:
    """Make a format string for the command to run.

    This interprets the execution_model and produces an appropriate shell command to
    start the implementation. This function produces commands for running locally:
    pinning is disabled and there's only one node.

    Args:
        implementation: The implementation to start.
        enable_debug: Whether to produce extra debug output.

    Return:
        A format string with embedded {ntasks} and {rankfile}.
    """
    if implementation.execution_model == ExecutionModel.DIRECT:
        fstr = 'exec {command} {args}'
    elif implementation.execution_model == ExecutionModel.OPENMPI:
        # Native name is orterun for older and prterun for newer OpenMPI.
        # So we go with mpirun, which works for either.
        fargs = [
                'exec mpirun -np $MUSCLE_MPI_PROCESSES',
                '--oversubscribe'
                ]

        if enable_debug:
            fargs.append('-v --debug-daemons --display-map --display-allocation')

        fargs.append('{command} {args}')

        fstr = ' '.join(fargs)

    elif implementation.execution_model == ExecutionModel.INTELMPI:
        fstr = 'exec mpirun -n $MUSCLE_MPI_PROCESSES {command} {args}'
    elif implementation.execution_model == ExecutionModel.SRUNMPI:
        raise ConfigurationError(
                f'Could not start {implementation.name} because the SRUNMPI execution'
                ' method only works in a SLURM allocation, and we are running locally.'
                ' Please switch this implementation to a different execution method'
                ' in the configuration file. You will probably want OPENMPI or'
                ' INTELMPI depending on which MPI implementation this code was'
                ' compiled with.')
    # elif implementation.execution_model == ExecutionModel.MPICH
    #    fstr = 'mpiexec -n {{ntasks}} {command} {args}'

    if implementation.args is None:
        args = ''
    elif isinstance(implementation.args, str):
        args = implementation.args
    elif isinstance(implementation.args, list):
        args = ' '.join(implementation.args)

    return fstr.format(
            command=implementation.executable,
            args=args
            )


def cluster_command(implementation: Implementation, enable_debug: bool) -> str:
    """Make a format string for the command to run.

    This interprets the execution_model and produces an appropriate shell command to
    start the implementation. This function produces commands for running on a cluster,
    with processes distributed across nodes and CPU pinning enabled.

    Args:
        implementation: The implementation to start.
        enable_debug: Whether to produce extra debug output.

    Return:
        A string with the command to use to start the implementation.
    """
    if implementation.execution_model == ExecutionModel.DIRECT:
        fargs = [
                'if ! taskset -V >/dev/null 2>&1 ; then',
                '    exec {command} {args}',
                'else',
                '    exec taskset $MUSCLE_BIND_MASK {command} {args}',
                'fi'
                ]
        fstr = '\n'.join(fargs)

    elif implementation.execution_model == ExecutionModel.OPENMPI:
        fargs = [
                # Native name is orterun for older and prterun for newer OpenMPI.
                # So we go with mpirun, which works for either.
                'exec mpirun -np $MUSCLE_MPI_PROCESSES',
                '--rankfile $MUSCLE_RANKFILE --use-hwthread-cpus --bind-to hwthread',
                '--oversubscribe'
                ]

        if enable_debug:
            fargs.append('-v --display-allocation --display-map --report-bindings')

        if slurm().quirks.overlap:
            # This adds the given option to the srun command used by mpirun to launch
            # its daemons. mpirun specifies --exclusive, which on SLURM <= 21-08 causes
            # SLURM to wait for our agents to quit, as it considers them to be occupying
            # the cores, causing a deadlock. Fortunately, it seems that adding --overlap
            # overrides the --exclusive and it works.
            fargs.append('-mca plm_slurm_args "--overlap"')

        fargs.append('{command} {args}')

        fstr = ' '.join(fargs)

    elif implementation.execution_model == ExecutionModel.INTELMPI:
        fargs = [
                'exec mpirun -n $MUSCLE_MPI_PROCESSES',
                '-machinefile $MUSCLE_RANKFILE']

        if enable_debug:
            fargs.append('-genv I_MPI_DEBUG=4')

        fargs.append('{command} {args}')

        fstr = ' '.join(fargs)

    elif implementation.execution_model == ExecutionModel.SRUNMPI:
        fargs = ['exec srun -n $MUSCLE_MPI_PROCESSES -m arbitrary']

        if slurm().quirks.overlap:
            fargs.append('--overlap')

        verbose = 'verbose,' if enable_debug else ''

        fargs.append(f'{slurm().quirks.cpu_bind}={verbose}$SLURM_CPU_BIND')
        fargs.append('{command} {args}')

        fstr = ' '.join(fargs)

    # elif implementation.execution_model == ExecutionModel.MPICH
    #    fstr = 'mpiexec -n $MUSCLE_MPI_PROCESSES -f $MUSCLE_RANKFILE {command} {args}'

    if implementation.args is None:
        args = ''
    elif isinstance(implementation.args, str):
        args = implementation.args
    elif isinstance(implementation.args, list):
        args = ' '.join(implementation.args)

    return fstr.format(
            command=implementation.executable,
            args=args
            )


def make_script(
        implementation: Implementation, res_req: ResourceRequirements,
        work_dir: Path, local: bool, rankfile: Optional[Path] = None) -> str:
    """Make a run script for a given implementation.

    Args:
        implementation: The implementation to launch
        res_req: The job's resource requirements
        work_dir: The directory to start the instance in
        local: Whether this is to run locally (True) or on a cluster (False)
        rankfile: Location of the rankfile, if any

    Return:
        A string with embedded newlines containing the shell script.
    """
    enable_debug = logging.getLogger('libmuscle').getEffectiveLevel() <= logging.DEBUG

    lines: List[str] = list()

    if implementation.base_env == BaseEnv.LOGIN:
        # We try to emulate an interactive login shell here by starting a
        # non-interactive login shell and manually loading the configuration files that
        # an interactive one would load. This avoids a confusing warning on stderr that
        # we get from /bin/bash -il because it can't find a controlling terminal.
        lines.append('#!/bin/bash --norc')
        lines.append('')
        lines.append('if [ -f /etc/environment ] ; then')
        lines.append('    . /etc/environment')
        lines.append('fi')
        lines.append('')
        lines.append('if [ -f /etc/profile ] ; then')
        lines.append('    . /etc/profile')
        lines.append('fi')
        lines.append('')
        lines.append('if [ -f ~/.bash_profile ] ; then')
        lines.append('    . ~/.bash_profile')
        lines.append('elif [ -f ~/.bash_login ] ; then')
        lines.append('    . ~/.bash_login')
        lines.append('elif [ -f ~/.profile ] ; then')
        lines.append('    . ~/.profile')
        lines.append('fi')
    else:
        lines.append('#!/bin/bash')

    lines.append('')

    # The environment is passed when starting the script, rather than as a set of
    # export statements here.

    if implementation.base_env == BaseEnv.CLEAN:
        lines.append('module purge')
        lines.append('')

    if implementation.modules:
        if isinstance(implementation.modules, str):
            lines.append(f'module load {implementation.modules}')
        else:
            for module in implementation.modules:
                lines.append(f'module load {module}')
        lines.append('')

    if implementation.virtual_env:
        lines.append(f'. {implementation.virtual_env}/bin/activate')
        lines.append('')

    if implementation.base_env == BaseEnv.LOGIN:
        # config files may change the work directory from what we set, so we add this to
        # ensure it's correct.
        lines.append(f'cd {work_dir}')

    if local:
        lines.append(local_command(implementation, enable_debug))
    else:
        lines.append(cluster_command(implementation, enable_debug))

    lines.append('')

    return '\n'.join(lines)



from typing import List, Optional, TypeVar

from ymmsl import Identifier, Operator
import ymmsl


_T = TypeVar("_T")


def _extend_list_to_size(lst: List[_T], size: int, padding: _T) -> None:
    """When lst is smaller than size, extend to size using padding as values
    """
    num_extend = size - len(lst)
    if num_extend > 0:
        lst += [padding] * num_extend


class Port(ymmsl.Port):
    """Represents a gateway to the outside world.

    Ports can be used to send or receive messages. They have a name and
    an operator, as well as a set of dimensions that determines the
    valid slot indices for sending or receiving on this port.

    Ports keep track of the amount of messages sent or received on the port.
    However, the actual incrementing and validation is done in
    :class:`Communicator`.

    Attributes:
        name (Identifier): Name of this port.
        operator (Operator): Operator associated with this port.
    """

    def __init__(self, name: str, operator: Operator, is_vector: bool,
                 is_connected: bool, our_ndims: int, peer_dims: List[int]
                 ) -> None:
        """Create a Port.

        Args:
            name: Name of this port.
            operator: Corresponding operator.
            is_vector: Whether this is a vector port.
            is_connected: Whether this port is connected to a peer.
            our_ndims: Number of dimensions of our instance set.
            peer_dims: Dimensions of the peer instance set of this port.
        """
        super().__init__(Identifier(name), operator)

        self._is_connected = is_connected

        if is_vector:
            if our_ndims == len(peer_dims):
                self._length: Optional[int] = 0
            elif our_ndims + 1 == len(peer_dims):
                self._length = peer_dims[-1]
            elif our_ndims > len(peer_dims):
                raise RuntimeError(('Vector port "{}" is connected to an'
                                    ' instance set with fewer dimensions.'
                                    ' It should be connected to a scalar'
                                    ' port on a set with one more dimension,'
                                    ' or to a vector port on a set with the'
                                    ' same number of dimensions.').format(
                                        name))
            else:
                raise RuntimeError(('Port "{}" is connected to an instance set'
                                    ' with more than one dimension more than'
                                    ' its own, which is not possible.').format(
                                        name))
            self._is_open = [True] * self._length
        else:
            if our_ndims < len(peer_dims):
                raise RuntimeError(('Scalar port "{}" is connected to an'
                                    ' instance set with more dimensions.'
                                    ' It should be connected to a scalar'
                                    ' port on an instance set with the same'
                                    ' dimensions, or to a vector port on an'
                                    ' instance set with one less dimension.'
                                    ).format(name))
            elif our_ndims > len(peer_dims) + 1:
                raise RuntimeError(('Scalar port "{}" is connected to an'
                                    ' instance set with at least two fewer'
                                    ' dimensions, which is not possible.'
                                    ).format(name))
            self._length = None
            self._is_open = [True]

        self._is_resizable = is_vector and (our_ndims == len(peer_dims))
        self._num_messages = [0] * (self._length or 1)
        self._is_resuming = [False] * (self._length or 1)

    # Note: I'm not sure how this will develop exactly, so this class has some
    # accessors even if those are un-Pythonic; in the future a simple variable
    # read may not be the right model.

    def is_connected(self) -> bool:
        """Returns whether the port is connected to a peer.

        Returns:
            True if there is a peer, False if there is not.
        """
        return self._is_connected

    def is_open(self, slot: Optional[int] = None) -> bool:
        """Returns whether this port is open.
        """
        if slot is not None:
            return self._is_open[slot]
        return self._is_open[0]

    def is_vector(self) -> bool:
        """Returns whether this is a vector port.

        Returns:
            True if it is vector, False if it is scalar.
        """
        return self._length is not None

    def is_resizable(self) -> bool:
        """Returns whether this port can be resized.
        """
        return self._is_resizable

    def get_length(self) -> int:
        """Returns the length of this port.

        Raises:
            RuntimeError: If this port is a scalar port.
        """
        if self._length is None:
            raise RuntimeError(('Tried to get length of scalar port {}'
                                ).format(self.name))
        return self._length

    def set_length(self, length: int) -> None:
        """Sets the length of a resizable vector port.

        Only call this if is_resizable() returns True.

        Args:
            length: The new length.

        Raises:
            RuntimeError: If the port is not resizable.
        """
        if not self._is_resizable:
            raise RuntimeError(('Tried to resize port {}, but it is not'
                                ' resizable.'.format(self.name)))
        if length != self._length:
            self._length = length
            self._is_open = [True] * self._length
            # Using extend here to not discard any information about message
            # numbers between resizes. Note that _num_messages and _is_resuming
            # may be longer than self._length!
            _extend_list_to_size(self._num_messages, self._length, 0)
            _extend_list_to_size(self._is_resuming, self._length, False)

    def set_closed(self, slot: Optional[int] = None) -> None:
        """Marks this port as closed.
        """
        if slot is not None:
            self._is_open[slot] = False
        else:
            self._is_open = [False]

    def restore_message_counts(self, num_messages: List[int]) -> None:
        """Restore message counts from a snapshot
        """
        self._num_messages = num_messages
        self._is_resuming = [True] * len(self._num_messages)
        _extend_list_to_size(self._num_messages, self._length or 1, 0)
        _extend_list_to_size(self._is_resuming, self._length or 1, False)

    def get_message_counts(self) -> List[int]:
        """Get a list of message counts for all slots in this port
        """
        return self._num_messages.copy()

    def increment_num_messages(self, slot: Optional[int] = None) -> None:
        """Increment amount of messages sent or received.

        Args:
            slot: The slot that is sent/received on
        """
        self._num_messages[slot or 0] += 1
        self.set_resumed(slot)

    def get_num_messages(self, slot: Optional[int] = None) -> int:
        """Get the amount of messages sent or received.

        Args:
            slot: The slot that is sent/received on
        """
        return self._num_messages[slot or 0]

    def is_resuming(self, slot: Optional[int] = None) -> bool:
        """True when this port has resumed.

        After resumption, each port/slot may discard exactly one message.
        is_resuming keeps track of this state.

        Args:
            slot: The slot that is sent/received on
        """
        return self._is_resuming[slot or 0]

    def set_resumed(self, slot: Optional[int] = None) -> None:
        """Mark that this port has resumed and may no longer discard messages.

        Args:
            slot: The slot that is sent/received on
        """
        self._is_resuming[slot or 0] = False



from threading import Condition
from typing import Dict, List, Set

from ymmsl import Port, Reference


class AlreadyRegistered(RuntimeError):
    pass


class InstanceRegistry:
    """Keeps track of running instances.

    The InstanceRegistry is a simple in-memory database that stores
    information about running instances of simulation components.
    """
    def __init__(self) -> None:
        """Construct an empty InstanceRegistry"""
        self._deregistered_one = Condition()    # doubles as lock
        self._locations: Dict[Reference, List[str]] = {}
        self._ports: Dict[Reference, List[Port]] = {}
        self._seen: Set[Reference] = set()
        self._startup = True

    def add(self, name: Reference, locations: List[str], ports: List[Port]
            ) -> None:
        """Add an instance to the registry.

        Args:
            name: Name of the instance.
            locations: Network locations where it can be reached.
            ports: List of ports of this instance.

        Raises:
            ValueError: If an instance with this name has already been
                    registered.
        """
        with self._deregistered_one:
            if name in self._locations:
                raise AlreadyRegistered(
                        'Instance {} tried to register twice'.format(name))

            self._locations[name] = locations
            self._ports[name] = ports
            self._seen.add(name)
            self._startup = False

    def get_locations(self, name: Reference) -> List[str]:
        """Retrieves the locations of a registered instance.

        Args:
            name: The name of the instance to get the location of.

        Raises:
            KeyError: If no instance with this name was registered.
        """
        with self._deregistered_one:
            return self._locations[name]

    def get_ports(self, name: Reference) -> List[Port]:
        """Retrieves the ports of a registered instance.

        Args:
            name: The name of the instance whose ports to retrieve.

        Raises:
            KeyError: If no instance with this name was registered.
        """
        with self._deregistered_one:
            return self._ports[name]

    def remove(self, name: Reference) -> None:
        """Remove an instance from the registry.

        Args:
            name: Name of the instance to remove.

        Raises:
            KeyError: If the instance does not exist.
        """
        with self._deregistered_one:
            del self._locations[name]
            del self._ports[name]
            self._deregistered_one.notify()

    def did_register(self, name: Reference) -> bool:
        """Check whether the instance has registered at some point.

        This returns True iff the instance registered at some point
        before the function was called. If it was subsequently
        deregistered, this will still return True.

        Args:
            name: Name of the instance to check.
        """
        return name in self._seen

    def wait(self) -> None:
        """Waits until all instances are deregistered.

        This function blocks, and returns after each instance has been
        registered and deregistered again, signally the end of the
        simulation run.
        """
        # This is called from a different thread than add/remove
        with self._deregistered_one:
            while self._startup or self._locations:
                self._deregistered_one.wait()



from typing import List

import msgpack
from ymmsl import Reference

from libmuscle.mcp.protocol import RequestType
from libmuscle.mcp.transport_server import RequestHandler, TransportServer
from libmuscle.mcp.type_registry import transport_server_types
from libmuscle.post_office import PostOffice


class MPPRequestHandler(RequestHandler):
    """Handles peer protocol requests.

    This accepts peer protocol message requests and responds to them by
    getting messages from a PostOffice.
    """
    def __init__(self, post_office: PostOffice) -> None:
        """Create an MPPRequestHandler.

        Args:
            post_office: The PostOffice to get messages from.
        """
        self._post_office = post_office

    def handle_request(self, request: bytes) -> bytes:
        """Handle a request.

        This receives an MCP request and handles it by blocking until
        the requested message is available, then returning it.

        Args:
            request: A received request

        Returns:
            An encoded response
        """
        req = msgpack.unpackb(request, raw=False)
        if len(req) != 2 or req[0] != RequestType.GET_NEXT_MESSAGE.value:
            raise RuntimeError(
                    'Invalid request type. Did the streams get crossed?')
        recv_port = Reference(req[1])
        return self._post_office.get_message(recv_port)


class MPPServer:
    """Serves MPP requests.

    This manages a collection of servers for different protocols and a
    PostOffice that stores outgoing messages.
    """
    def __init__(self) -> None:
        self._post_office = PostOffice()
        self._handler = MPPRequestHandler(self._post_office)
        self._servers: List[TransportServer] = []

        for server_type in transport_server_types:
            server = server_type(self._handler)
            self._servers.append(server)

    def get_locations(self) -> List[str]:
        """Returns a list of locations that we can be reached at.

        These locations are of the form 'protocol:location', where
        the protocol name does not contain a colon and location may
        be an arbitrary string.

        Returns:
            A list of strings describing network locations.
        """
        return [server.get_location() for server in self._servers]

    def deposit(self, receiver: Reference, message: bytes) -> None:
        """Deposits a message for the receiver to retrieve.

        Args:
            receiver: Receiver of the message.
            message: The message to deposit.
        """
        self._post_office.deposit(receiver, message)

    def wait_for_receivers(self) -> None:
        """Waits for all deposited messages to have been received."""
        self._post_office.wait_for_receivers()

    def shutdown(self) -> None:
        """Shut down all servers."""
        for server in self._servers:
            server.close()



from pathlib import Path
from unittest.mock import MagicMock

from ymmsl import Reference, Settings

from libmuscle.communicator import Message
from libmuscle.snapshot import SnapshotMetadata
from libmuscle.snapshot_manager import SnapshotManager


def test_no_checkpointing(tmp_path: Path) -> None:
    manager = MagicMock()
    port_manager = MagicMock()
    port_manager.get_message_counts.return_value = {}
    snapshot_manager = SnapshotManager(Reference('test'), manager, port_manager)

    snapshot_manager.prepare_resume(None, tmp_path)
    assert not snapshot_manager.resuming_from_intermediate()
    assert not snapshot_manager.resuming_from_final()


def test_save_load_snapshot(tmp_path: Path) -> None:
    manager = MagicMock()
    port_manager = MagicMock()
    port_message_counts = {'in': [1], 'out': [2], 'muscle_settings_in': [0]}
    port_manager.get_message_counts.return_value = port_message_counts

    instance_id = Reference('test[1]')
    snapshot_manager = SnapshotManager(instance_id, manager, port_manager)

    snapshot_manager.prepare_resume(None, tmp_path)
    assert not snapshot_manager.resuming_from_intermediate()
    assert not snapshot_manager.resuming_from_final()

    snapshot_manager.save_snapshot(
            Message(0.2, None, 'test data'), False, ['test'], 13.0, None,
            Settings())

    port_manager.get_message_counts.assert_called_with()
    manager.submit_snapshot_metadata.assert_called()
    metadata, = manager.submit_snapshot_metadata.call_args[0]
    assert isinstance(metadata, SnapshotMetadata)
    assert metadata.triggers == ['test']
    assert metadata.wallclock_time == 13.0
    assert metadata.timestamp == 0.2
    assert metadata.next_timestamp is None
    assert metadata.port_message_counts == port_message_counts
    assert not metadata.is_final_snapshot
    snapshot_path = Path(metadata.snapshot_filename)
    assert snapshot_path.parent == tmp_path
    assert snapshot_path.name == 'test-1_1.pack'

    snapshot_manager2 = SnapshotManager(instance_id, manager, port_manager)

    snapshot_manager2.prepare_resume(snapshot_path, tmp_path)
    port_manager.restore_message_counts.assert_called_with(port_message_counts)

    assert snapshot_manager2.resuming_from_intermediate()
    assert not snapshot_manager2.resuming_from_final()
    msg = snapshot_manager2.load_snapshot()
    assert msg.timestamp == 0.2
    assert msg.next_timestamp is None
    assert msg.data == 'test data'

    snapshot_manager2.save_snapshot(
            Message(0.6, None, 'test data2'), True, ['test'], 42.2, 1.2,
            Settings())

    metadata, = manager.submit_snapshot_metadata.call_args[0]
    assert isinstance(metadata, SnapshotMetadata)
    assert metadata.triggers == ['test']
    assert metadata.wallclock_time == 42.2
    assert metadata.timestamp == 0.6
    assert metadata.next_timestamp is None
    assert metadata.port_message_counts == port_message_counts
    assert metadata.is_final_snapshot
    snapshot_path = Path(metadata.snapshot_filename)
    assert snapshot_path.parent == tmp_path
    assert snapshot_path.name == 'test-1_3.pack'


def test_save_load_implicit_snapshot(tmp_path: Path) -> None:
    manager = MagicMock()
    port_manager = MagicMock()
    port_message_counts = {'in': [1], 'out': [2], 'muscle_settings_in': [0]}
    port_manager.get_message_counts.return_value = port_message_counts

    instance_id = Reference('test[1]')
    snapshot_manager = SnapshotManager(instance_id, manager, port_manager)

    snapshot_manager.prepare_resume(None, tmp_path)

    assert not snapshot_manager.resuming_from_intermediate()
    assert not snapshot_manager.resuming_from_final()
    # save implicit snapshot
    snapshot_manager.save_snapshot(
            None, True, ['implicit'], 1.0, 1.5, Settings())

    manager.submit_snapshot_metadata.assert_called_once()
    metadata, = manager.submit_snapshot_metadata.call_args[0]
    assert isinstance(metadata, SnapshotMetadata)
    snapshot_path = Path(metadata.snapshot_filename)
    manager.submit_snapshot_metadata.reset_mock()

    snapshot_manager2 = SnapshotManager(instance_id, manager, port_manager)

    snapshot_manager2.prepare_resume(snapshot_path, tmp_path)
    port_manager.restore_message_counts.assert_called_with(port_message_counts)
    manager.submit_snapshot_metadata.assert_called_once()
    manager.submit_snapshot_metadata.reset_mock()

    assert not snapshot_manager2.resuming_from_intermediate()
    assert not snapshot_manager2.resuming_from_final()
    snapshot_manager2.save_snapshot(
            None, True, ['implicit'], 12.3, 2.5, Settings())
    manager.submit_snapshot_metadata.assert_called_once()



from typing import cast, Dict, List, Tuple

from ymmsl import Conduit, Identifier, Reference

from libmuscle.endpoint import Endpoint


class PeerInfo:
    """Interprets information about peers for a Communicator
    """
    def __init__(self, kernel: Reference, index: List[int],
                 conduits: List[Conduit],
                 peer_dims: Dict[Reference, List[int]],
                 peer_locations: Dict[Reference, List[str]]) -> None:
        """Create a PeerInfo.

        Peers here are instances, and peer_dims and peer_locations are
        indexed by a Reference to an instance. Instance sets are
        multi-dimensional arrays with sizes given by peer_dims.

        Args:
            kernel: The kernel for the instance whose peers we're
                    managing.
            index: The index of the instance whose peers we're
                    managing.
            conduits: A list of conduits attached to this component,
                    as received from the manager.
            peer_dims: For each peer we share a conduit with, the
                    dimensions of the instance set.
            peer_locations: A list of locations for each peer instance
                    we share a conduit with.
        """
        self.__kernel = kernel
        self.__index = index

        self._incoming_ports: List[Reference] = []
        self._outgoing_ports: List[Reference] = []

        # peer port ids, indexed by local kernel.port id
        self.__peers: Dict[Reference, List[Reference]] = {}

        for conduit in conduits:
            if str(conduit.sending_component()) == str(kernel):
                # we send on the port this conduit attaches to
                if conduit.sender not in self._outgoing_ports:
                    self._outgoing_ports.append(conduit.sender)
                self.__peers.setdefault(conduit.sender, []).append(
                        conduit.receiver)

            if str(conduit.receiving_component()) == str(kernel):
                # we receive on the port this conduit attaches to
                if conduit.receiver in self.__peers:
                    raise RuntimeError(('Receiving port "{}" is connected by'
                                        ' multiple conduits, but at most one'
                                        ' is allowed.'
                                        ).format(conduit.receiving_port()))
                self._incoming_ports.append(conduit.receiver)
                self.__peers[conduit.receiver] = [conduit.sender]

        self.__peer_dims = peer_dims    # indexed by kernel id
        self.__peer_locations = peer_locations  # indexed by instance id

    def list_incoming_ports(self) -> List[Tuple[Identifier, Reference]]:
        """List incoming ports.

        Returns:
            A list of tuples containing a port id and a reference to the
            peer endpoint.
        """
        return [
                (cast(Identifier, port_ref[-1]), self.__peers[port_ref][0])
                for port_ref in self._incoming_ports]

    def list_outgoing_ports(self) -> List[Tuple[Identifier, List[Reference]]]:
        """List outgoing ports.

        Returns:
            A list of tuples containing a port id and a list of references
            to the peer endpoint(s).
        """
        return [
                (cast(Identifier, port_ref[-1]), self.__peers[port_ref])
                for port_ref in self._outgoing_ports]

    def is_connected(self, port: Identifier) -> bool:
        """Determine whether the given port is connected.

        Args:
            port: The port to check.
        """
        recv_port_full = self.__kernel + port
        return recv_port_full in self.__peers

    def get_peer_ports(self, port: Identifier) -> List[Reference]:
        """Get a reference for the peer ports.

        Args:
            port: Name of the port on this side.
        """
        return self.__peers[self.__kernel + port]

    def get_peer_dims(self, peer_kernel: Reference) -> List[int]:
        """Get the dimensions of a peer kernel.

        Args:
            peer_kernel: The peer kernel whose dimensions to get.
        """
        return self.__peer_dims[peer_kernel]

    def get_peer_locations(self, peer_instance: Reference) -> List[str]:
        """Get the locations of a peer instance.

        There may be multiple, if the peer supports more than one
                protocol.

        Args:
            peer_instance: The instance whose locations to get.
        """
        return self.__peer_locations[peer_instance]

    def get_peer_endpoints(self, port: Identifier, slot: List[int]
                           ) -> List[Endpoint]:
        """Determine the peer endpoints for the given port and slot.

        Args:
            port: The port on our side to send or receive on.
            slot: The slot to send or receive on.

        Returns:
            The peer endpoints.
        """
        peers = self.__peers[self.__kernel + port]
        endpoints = []

        for peer in peers:
            peer_kernel = peer[:-1]
            peer_port = cast(Identifier, peer[-1])

            total_index = self.__index + slot

            # rebalance the indices
            peer_dim = len(self.__peer_dims[peer_kernel])
            peer_index = total_index[0:peer_dim]
            peer_slot = total_index[peer_dim:]
            endpoints.append(
                    Endpoint(peer_kernel, peer_index, peer_port, peer_slot))

        return endpoints



import dataclasses
from pathlib import Path
from random import uniform
from threading import Lock
from time import perf_counter, sleep
from typing import Any, Dict, Iterable, List, Optional, Tuple

import msgpack
from ymmsl import (
        Conduit, Operator, Port, Reference, Settings, Checkpoints,
        CheckpointRule, CheckpointRangeRule, CheckpointAtRule)

import libmuscle
from libmuscle.mcp.protocol import RequestType, ResponseType
from libmuscle.mcp.tcp_transport_client import TcpTransportClient
from libmuscle.profiling import ProfileEvent
from libmuscle.logging import LogMessage
from libmuscle.snapshot import SnapshotMetadata


PEER_TIMEOUT = 600
PEER_INTERVAL_MIN = 5.0
PEER_INTERVAL_MAX = 10.0

_CheckpointInfoType = Tuple[
        float, Checkpoints, Optional[Path], Optional[Path]]


def encode_operator(op: Operator) -> str:
    """Convert an Operator to a MsgPack-compatible value."""
    return op.name


def encode_port(port: Port) -> List[str]:
    """Convert a Port to a MsgPack-compatible value."""
    return [str(port.name), encode_operator(port.operator)]


def encode_profile_event(event: ProfileEvent) -> Any:
    """Converts a ProfileEvent to a list.

    Args:
        event: A profile event

    Returns:
        A list with its attributes, for MMP serialisation.
    """
    if event.start_time is None or event.stop_time is None:
        raise RuntimeError(
                'Incomplete ProfileEvent sent. This is a bug, please'
                ' report it.')

    encoded_port = encode_port(event.port) if event.port else None
    return [
            event.event_type.value,
            event.start_time.nanoseconds, event.stop_time.nanoseconds,
            encoded_port, event.port_length, event.slot,
            event.message_number, event.message_size, event.message_timestamp]


def decode_checkpoint_rule(rule: Dict[str, Any]) -> CheckpointRule:
    """Decode a checkpoint rule from a MsgPack-compatible value."""
    if rule.keys() == {'at'}:
        return CheckpointAtRule(**rule)
    if rule.keys() == {'start', 'stop', 'every'}:
        return CheckpointRangeRule(**rule)
    raise ValueError(f'Cannot convert {rule} to a checkpoint rule.')


def decode_checkpoint_info(
        elapsed_time: float,
        checkpoints_dict: Dict[str, Any],
        resume: Optional[str],
        snapshot_dir: Optional[str]
        ) -> _CheckpointInfoType:
    """Decode checkpoint info from a MsgPack-compatible value.

    Args:
        elapsed_time: current elapsed time according to the manager
        checkpoints_dict: checkpoint definitions from the MsgPack
        resume: path to the snapshot we should resume from, if any
        snapshot_dir: path to the directory to store new snapshots in

    Returns:
        elapsed_time: current elapsed time according to the manager
        checkpoints: checkpoint configuration
        resume: path to the snapshot we should resume from, if any
        snapshot_dir: path to the directory to store new snapshots in
    """
    checkpoints = Checkpoints(
            at_end=checkpoints_dict["at_end"],
            wallclock_time=[decode_checkpoint_rule(rule)
                            for rule in checkpoints_dict["wallclock_time"]],
            simulation_time=[decode_checkpoint_rule(rule)
                             for rule in checkpoints_dict["simulation_time"]])
    resume_path = None if resume is None else Path(resume)
    snapshot_path = None if snapshot_dir is None else Path(snapshot_dir)
    return (elapsed_time, checkpoints, resume_path, snapshot_path)


class MMPClient():
    """The client for the MUSCLE Manager Protocol.

    This class connects to the Manager and communicates with it on
    behalf of the rest of libmuscle.

    It manages the connection, and converts between our native types
    and the gRPC generated types.

    Communication is protected by an internal lock, so this class can
    be called simultaneously from different threads.
    """
    def __init__(self, instance_id: Reference, location: str) -> None:
        """Create an MMPClient

        Args:
            location: A connection string of the form hostname:port
        """
        self._instance_id = instance_id
        self._transport_client = TcpTransportClient(location)
        self._mutex = Lock()

    def close(self) -> None:
        """Close the connection

        This closes the connection. After this no other member
        functions can be called.
        """
        self._transport_client.close()

    def submit_log_message(self, message: LogMessage) -> None:
        """Send a log message to the manager.

        Args:
            message: The message to send.
        """
        request = [
                RequestType.SUBMIT_LOG_MESSAGE.value,
                message.instance_id, message.timestamp.seconds,
                message.level.value, message.text]
        self._call_manager(request)

    def submit_profile_events(self, events: Iterable[ProfileEvent]) -> None:
        """Sends profiling events to the manager.

        Args:
            events: The events to send.
        """
        request = [
                RequestType.SUBMIT_PROFILE_EVENTS.value,
                str(self._instance_id),
                [encode_profile_event(e) for e in events]]
        self._call_manager(request)

    def submit_snapshot_metadata(
            self, snapshot_metadata: SnapshotMetadata) -> None:
        """Send snapshot metadata to the manager.

        Args:
            snapshot_metadata: Snapshot metadata to supply to the manager.
        """
        request = [
                RequestType.SUBMIT_SNAPSHOT.value,
                str(self._instance_id),
                dataclasses.asdict(snapshot_metadata)]
        self._call_manager(request)

    def get_settings(self) -> Settings:
        """Get the central settings from the manager.

        Returns:
            The requested settings.
        """
        request = [RequestType.GET_SETTINGS.value]
        response = self._call_manager(request)
        return Settings(response[1])

    def get_checkpoint_info(self) -> _CheckpointInfoType:
        """Get the checkpoint info from the manager.

        Returns:
            elapsed_time: current elapsed time
            checkpoints: checkpoint configuration
            resume: path to the resume snapshot
            snapshot_directory: path to store snapshots
        """
        request = [RequestType.GET_CHECKPOINT_INFO.value, str(self._instance_id)]
        response = self._call_manager(request)
        return decode_checkpoint_info(*response[1:])

    def register_instance(
            self, locations: List[str], ports: List[Port]) -> None:
        """Register a component instance with the manager.

        Args:
            locations: List of places where the instance can be
                    reached.
            ports: List of ports of this instance.
        """
        request = [
                RequestType.REGISTER_INSTANCE.value,
                str(self._instance_id), locations,
                [encode_port(p) for p in ports],
                libmuscle.__version__]
        response = self._call_manager(request)
        if response[0] == ResponseType.ERROR.value:
            raise RuntimeError(
                    f'Error registering instance: {response[1]}')

    def request_peers(self) -> Tuple[
            List[Conduit],
            Dict[Reference, List[int]],
            Dict[Reference, List[str]]]:
        """Request connection information about peers.

        This will repeat the request at an exponentially increasing
        query interval at first, until it reaches the interval
        specified by PEER_INTERVAL_MIN and PEER_INTERVAL_MAX. From
        there on, intervals are drawn randomly from that range.

        Returns:
            A tuple containing a list of conduits that this instance is
            attached to, a dictionary of peer dimensions, which is
            indexed by Reference to the peer kernel, and specifies how
            many instances of that kernel there are, and a dictionary
            of peer instance locations, indexed by Reference to a peer
            instance, and containing for each peer instance a list of
            network location strings at which it can be reached.
        """
        sleep_time = 0.1
        start_time = perf_counter()

        request = [RequestType.GET_PEERS.value, str(self._instance_id)]
        response = self._call_manager(request)

        while (response[0] == ResponseType.PENDING.value and
               perf_counter() < start_time + PEER_TIMEOUT and
               sleep_time < PEER_INTERVAL_MIN):
            sleep(sleep_time)
            response = self._call_manager(request)
            sleep_time *= 1.5

        while (response[0] == ResponseType.PENDING.value and
               perf_counter() < start_time + PEER_TIMEOUT):
            sleep(uniform(PEER_INTERVAL_MIN, PEER_INTERVAL_MAX))
            response = self._call_manager(request)

        if response[0] == ResponseType.PENDING.value:
            raise RuntimeError('Timeout waiting for peers to appear')

        if response[0] == ResponseType.ERROR.value:
            raise RuntimeError('Error getting peers from manager: {}'.format(
                    response[1]))

        conduits = [Conduit(snd, recv) for snd, recv in response[1]]

        peer_dimensions = {
                Reference(component): dims
                for component, dims in response[2].items()}

        peer_locations = {
                Reference(instance): locs
                for instance, locs in response[3].items()}

        return conduits, peer_dimensions, peer_locations

    def deregister_instance(self) -> None:
        """Deregister a component instance with the manager.
        """
        request = [
                RequestType.DEREGISTER_INSTANCE.value, str(self._instance_id)]
        response = self._call_manager(request)

        if response[0] == ResponseType.ERROR.value:
            raise RuntimeError('Error deregistering instance: {}'.format(
                    response[1]))

    def waiting_for_receive(
            self, peer_instance_id: Reference, port_name: str, slot: Optional[int]
            ) -> None:
        """Notify the manager that we're waiting to receive a message."""
        request = [
                RequestType.WAITING_FOR_RECEIVE.value,
                str(self._instance_id),
                str(peer_instance_id), port_name, slot]
        self._call_manager(request)

    def waiting_for_receive_done(
            self, peer_instance_id: Reference, port_name: str, slot: Optional[int]
            ) -> None:
        """Notify the manager that we're done waiting to receive a message."""
        request = [
                RequestType.WAITING_FOR_RECEIVE_DONE.value,
                str(self._instance_id),
                str(peer_instance_id), port_name, slot]
        self._call_manager(request)

    def is_deadlocked(self) -> bool:
        """Ask the manager if we're part of a deadlock."""
        request = [
                RequestType.IS_DEADLOCKED.value, str(self._instance_id)]
        response = self._call_manager(request)
        return bool(response[1])

    def _call_manager(self, request: Any) -> Any:
        """Call the manager and do en/decoding.

        Args:
            request: The request to encode and send

        Returns:
            The decoded response
        """
        with self._mutex:
            encoded_request = msgpack.packb(request, use_bin_type=True)
            response, _ = self._transport_client.call(encoded_request)
            return msgpack.unpackb(response, raw=False)



import socket
import socketserver as ss
import threading
from typing import cast, List, Optional, Tuple, Type

import psutil

from libmuscle.mcp.transport_server import RequestHandler, TransportServer
from libmuscle.mcp.tcp_util import (recv_all, recv_int64, send_int64,
                                    SocketClosed)


class TcpTransportServerImpl(ss.ThreadingMixIn, ss.TCPServer):
    daemon_threads = True
    allow_reuse_address = True

    def __init__(self, host_port_tuple: Tuple[str, int],
                 streamhandler: Type, transport_server: 'TcpTransportServer'
                 ) -> None:
        super().__init__(host_port_tuple, streamhandler)
        self.transport_server = transport_server
        if hasattr(socket, "TCP_NODELAY"):
            self.socket.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        if hasattr(socket, "TCP_QUICKACK"):
            self.socket.setsockopt(socket.SOL_TCP, socket.TCP_QUICKACK, 1)


class TcpHandler(ss.BaseRequestHandler):
    """Handler for MCP-over-TCP connections.

    This is a Python handler for Python's TCPServer, which forwards
    to the RequestHandler attached to the server.
    """
    def handle(self) -> None:
        """Handles requests on a socket
        """
        request = self.receive_request()

        while request is not None:
            server = cast(TcpTransportServerImpl, self.server).transport_server
            response = server._handler.handle_request(request)

            send_int64(self.request, len(response))
            self.request.sendall(response)
            request = self.receive_request()

    def receive_request(self) -> Optional[bytes]:
        """Receives a request

        Returns:
            The received bytes
        """
        try:
            length = recv_int64(self.request)
            reqbuf = recv_all(self.request, length)
            return reqbuf
        except SocketClosed:
            return None

    def finish(self) -> None:
        """Called when shutting down the thread?"""
        server = cast(TcpTransportServerImpl, self.server).transport_server
        server._handler.close()


class TcpTransportServer(TransportServer):
    """A TransportServer that uses TCP to communicate."""
    def __init__(self, handler: RequestHandler, port: int = 0) -> None:
        """Create a TCPServer.

        Args:
            handler: A RequestHandler to handle requests
            port: The port to use.

        Raises:
            OSError: With errno set to errno.EADDRINUSE if the port is not
                available.
        """
        super().__init__(handler)

        self._server = TcpTransportServerImpl(('', port), TcpHandler, self)
        self._server_thread = threading.Thread(
                target=self._server.serve_forever, args=(0.1,), daemon=True)
        self._server_thread.start()

    def get_location(self) -> str:
        """Returns the location this server listens on.

        Returns:
            A string containing the location.
        """
        host, port = self._server.server_address

        locs: List[str] = []
        for address in self._get_if_addresses():
            locs.append('{}:{}'.format(address, port))
        return 'tcp:{}'.format(','.join(locs))

    def close(self) -> None:
        """Closes this server.

        Stops the server listening, waits for existing clients to
        disconnect, then frees any other resources.
        """
        self._server.shutdown()
        self._server_thread.join()
        self._server.server_close()

    def _get_if_addresses(self) -> List[str]:
        """Returns a list of local addresses.

        This returns a list of strings containing all IPv4 and IPv6 network
        addresses bound to the available network interfaces. The server
        will listen on all interfaces, but not all of them may be reachable
        from the client. So we get all of them here, and the client can
        then try them all and find one that works.
        """
        all_addresses: List[str] = []
        ifs = psutil.net_if_addrs()
        for _, addresses in ifs.items():
            for addr in addresses:
                if addr.family == socket.AF_INET:
                    if not addr.address.startswith('127.'):
                        all_addresses.append(addr.address)
                if addr.family == socket.AF_INET6:
                    # filter out link-local addresses with a scope id
                    if '%' not in addr.address and addr.address != '::1':
                        all_addresses.append('[' + addr.address + ']')

        return all_addresses



import logging
from typing import Any, Dict, List, Optional, Tuple, cast
from ymmsl import Identifier, Reference, Settings

from libmuscle.endpoint import Endpoint
from libmuscle.mmp_client import MMPClient
from libmuscle.mpp_message import ClosePort, MPPMessage
from libmuscle.mpp_client import MPPClient
from libmuscle.mpp_server import MPPServer
from libmuscle.mcp.tcp_util import SocketClosed
from libmuscle.peer_info import PeerInfo
from libmuscle.port_manager import PortManager
from libmuscle.profiler import Profiler
from libmuscle.profiling import (
        ProfileEvent, ProfileEventType, ProfileTimestamp)
from libmuscle.receive_timeout_handler import Deadlock, ReceiveTimeoutHandler


_logger = logging.getLogger(__name__)


MessageObject = Any


class Message:
    """A message to be sent or received.

    This class describes a message to be sent or that has been
    received.

    Attributes:
        timestamp (float): Simulation time for which this data is valid.
        next_timestamp (Optional[float]): Simulation time for the next
                message to be transmitted through this port.
        data (MessageObject): An object to send or that was received.
        settings (Settings): Overlay settings to send or that was
                received.
    """
    # Note: This is for communication with the user, it's not what
    # actually goes out on the wire, see libmuscle.mcp.Message for that.
    def __init__(self, timestamp: float, next_timestamp: Optional[float] = None,
                 data: MessageObject = None,
                 settings: Optional[Settings] = None
                 ) -> None:
        """Create a Message.

        Args:
            timestamp: Simulation time for which this data is valid.
            next_timestamp: Simulation time for the next message to be
                    transmitted through this port.
            data: An object to send or that was received.
            settings: Overlay settings to send or that were received.
        """
        # make sure timestamp and next_timestamp are floats
        timestamp = float(timestamp)
        if next_timestamp is not None:
            next_timestamp = float(next_timestamp)

        self.timestamp = timestamp
        self.next_timestamp = next_timestamp
        self.data = data
        self.settings = settings


class Communicator:
    """Communication engine for MUSCLE3.

    This class is the mailroom for a kernel that uses MUSCLE3. It
    manages the sending and receiving of messages, although it
    leaves the actual data transmission to various protocol-specific
    servers and clients.
    """
    def __init__(
            self, kernel: Reference, index: List[int],
            port_manager: PortManager, profiler: Profiler,
            manager: MMPClient) -> None:
        """Create a Communicator.

        The instance reference must start with one or more Identifiers,
        giving the kernel id, followed by one or more integers which
        specify the instance index.

        Args:
            kernel: The kernel this is the Communicator for.
            index: The index for this instance.
            port_manager: The PortManager to use.
            profiler: The profiler to use for recording sends and
                    receives.
        """
        self._kernel = kernel
        self._index = index
        self._port_manager = port_manager
        self._profiler = profiler
        self._manager = manager
        # Notify manager, by default, after 10 seconds waiting in receive_message()
        self._receive_timeout = 10.0

        self._server = MPPServer()

        # indexed by remote instance id
        self._clients: Dict[Reference, MPPClient] = {}

    def get_locations(self) -> List[str]:
        """Returns a list of locations that we can be reached at.

        These locations are of the form 'protocol:location', where
        the protocol name does not contain a colon and location may
        be an arbitrary string.

        Returns:
            A list of strings describing network locations.
        """
        return self._server.get_locations()

    def set_peer_info(self, peer_info: PeerInfo) -> None:
        """Inform this Communicator about its peers.

        This tells the Communicator about its peers, so that it can route
        messages accordingly.

        Args:
            peer_info: Information about the peers.
        """
        self._peer_info = peer_info

    def set_receive_timeout(self, receive_timeout: float) -> None:
        """Update the timeout after which the manager is notified that we are waiting
        for a message.

        Args:
            receive_timeout: Timeout (seconds). A negative number disables the deadlock
                notification mechanism.
        """
        self._receive_timeout = receive_timeout

    def send_message(
            self, port_name: str, message: Message,
            slot: Optional[int] = None,
            checkpoints_considered_until: float = float('-inf')) -> None:
        """Send a message and settings to the outside world.

        Sending is non-blocking, a copy of the message will be made
        and stored until the receiver is ready to receive it.

        Args:
            port_name: The port on which this message is to be sent.
            message: The message to be sent.
            slot: The slot to send the message on, if any.
            checkpoints_considered_until: When we last checked if we
                should save a snapshot (wallclock time).
        """
        if slot is None:
            _logger.debug('Sending message on {}'.format(port_name))
            slot_list: List[int] = []
        else:
            _logger.debug('Sending message on {}[{}]'.format(port_name, slot))
            slot_list = [slot]

        snd_endpoint = self.__get_endpoint(port_name, slot_list)
        if not self._port_manager.get_port(str(snd_endpoint.port)).is_connected():
            # log sending on disconnected port
            return

        port = self._port_manager.get_port(port_name)
        profile_event = ProfileEvent(
                ProfileEventType.SEND, ProfileTimestamp(), None, port, None,
                slot, port.get_num_messages(slot), None, message.timestamp)

        recv_endpoints = self._peer_info.get_peer_endpoints(
                snd_endpoint.port, slot_list)

        port_length = None
        if port.is_resizable():
            port_length = port.get_length()

        for recv_endpoint in recv_endpoints:
            mpp_message = MPPMessage(snd_endpoint.ref(), recv_endpoint.ref(),
                                     port_length,
                                     message.timestamp, message.next_timestamp,
                                     cast(Settings, message.settings),
                                     port.get_num_messages(slot),
                                     checkpoints_considered_until,
                                     message.data)
            encoded_message = mpp_message.encoded()
            self._server.deposit(recv_endpoint.ref(), encoded_message)

        port.increment_num_messages(slot)

        profile_event.stop()
        if port.is_vector():
            profile_event.port_length = port.get_length()
        profile_event.message_size = len(encoded_message)
        if not isinstance(message.data, ClosePort):
            self._profiler.record_event(profile_event)

    def receive_message(
            self, port_name: str, slot: Optional[int] = None) -> Tuple[Message, float]:
        """Receive a message and attached settings overlay.

        Receiving is a blocking operation. This function will contact
        the sender, wait for a message to be available, and receive and
        return it.

        If the port is not connected, then the default value will be
        returned if one was given, exactly as it was given. If no
        default was given then a RuntimeError will be raised.

        Args:
            port_name: The endpoint on which a message is to be
                    received.
            slot: The slot to receive the message on, if any.

        Returns:
            The received message, with message.settings holding
            the settings overlay. The settings attribute is
            guaranteed to not be None. Secondly, the saved_until
            metadata field from the received message.

        Raises:
            RuntimeError: If the network connection had an error, or the
                    message number was incorrect.
        """
        if port_name == 'muscle_settings_in':
            port = self._port_manager._muscle_settings_in
        else:
            port = self._port_manager.get_port(port_name)

        if slot is None:
            port_and_slot = port_name
            slot_list: List[int] = []
        else:
            port_and_slot = f"{port_name}[{slot}]"
            slot_list = [slot]
        _logger.debug('Waiting for message on {}'.format(port_and_slot))

        recv_endpoint = self.__get_endpoint(port_name, slot_list)

        receive_event = ProfileEvent(
                ProfileEventType.RECEIVE, ProfileTimestamp(), None, port, None,
                slot, port.get_num_messages())

        # peer_info already checks that there is at most one snd_endpoint
        # connected to the port we receive on
        snd_endpoint = self._peer_info.get_peer_endpoints(
                recv_endpoint.port, slot_list)[0]
        client = self.__get_client(snd_endpoint.instance())
        timeout_handler = None
        if self._receive_timeout >= 0:
            timeout_handler = ReceiveTimeoutHandler(
                    self._manager, snd_endpoint.instance(),
                    port_name, slot, self._receive_timeout)
        try:
            mpp_message_bytes, profile = client.receive(
                    recv_endpoint.ref(), timeout_handler)
        except (ConnectionError, SocketClosed) as exc:
            raise RuntimeError(
                "Error while receiving a message: connection with peer"
                f" '{snd_endpoint.kernel}' was lost. Did the peer crash?"
            ) from exc
        except Deadlock:
            # Profiler messages may be used for debugging the deadlock
            self._profiler.shutdown()
            raise RuntimeError(
                "Deadlock detected while receiving a message on "
                f"port '{port_and_slot}'. See manager logs for more detail."
            ) from None

        recv_decode_event = ProfileEvent(
                ProfileEventType.RECEIVE_DECODE, ProfileTimestamp(), None,
                port, None, slot, port.get_num_messages(),
                len(mpp_message_bytes))
        mpp_message = MPPMessage.from_bytes(mpp_message_bytes)
        recv_decode_event.stop()

        if mpp_message.port_length is not None:
            if port.is_resizable():
                port.set_length(mpp_message.port_length)

        if isinstance(mpp_message.data, ClosePort):
            port.set_closed(slot)

        message = Message(
                mpp_message.timestamp, mpp_message.next_timestamp,
                mpp_message.data, mpp_message.settings_overlay)

        recv_wait_event = ProfileEvent(
                ProfileEventType.RECEIVE_WAIT, profile[0], profile[1], port,
                mpp_message.port_length, slot, port.get_num_messages(),
                len(mpp_message_bytes), message.timestamp)

        recv_xfer_event = ProfileEvent(
                ProfileEventType.RECEIVE_TRANSFER, profile[1], profile[2],
                port, mpp_message.port_length, slot, port.get_num_messages(),
                len(mpp_message_bytes), message.timestamp)

        recv_decode_event.message_timestamp = message.timestamp
        receive_event.message_timestamp = message.timestamp

        if port.is_vector():
            receive_event.port_length = port.get_length()
            recv_wait_event.port_length = port.get_length()
            recv_xfer_event.port_length = port.get_length()
            recv_decode_event.port_length = port.get_length()

        receive_event.message_size = len(mpp_message_bytes)

        if not isinstance(mpp_message.data, ClosePort):
            self._profiler.record_event(recv_wait_event)
            self._profiler.record_event(recv_xfer_event)
            self._profiler.record_event(recv_decode_event)
            self._profiler.record_event(receive_event)

        expected_message_number = port.get_num_messages(slot)
        if expected_message_number != mpp_message.message_number:
            if (expected_message_number - 1 == mpp_message.message_number and
                    port.is_resuming(slot)):
                _logger.debug(f'Discarding received message on {port_and_slot}'
                              ': resuming from weakly consistent snapshot')
                port.set_resumed(slot)
                return self.receive_message(port_name, slot)
            raise RuntimeError(f'Received message on {port_and_slot} with'
                               ' unexpected message number'
                               f' {mpp_message.message_number}. Was expecting'
                               f' {expected_message_number}. Are you resuming'
                               ' from an inconsistent snapshot?')
        port.increment_num_messages(slot)

        _logger.debug('Received message on {}'.format(port_and_slot))
        if isinstance(mpp_message.data, ClosePort):
            _logger.debug('Port {} is now closed'.format(port_and_slot))

        return message, mpp_message.saved_until

    def shutdown(self) -> None:
        """Shuts down the Communicator, closing connections.
        """
        self._close_ports()

        for client in self._clients.values():
            client.close()

        wait_event = ProfileEvent(ProfileEventType.DISCONNECT_WAIT, ProfileTimestamp())
        self._server.wait_for_receivers()
        self._profiler.record_event(wait_event)

        shutdown_event = ProfileEvent(ProfileEventType.SHUTDOWN, ProfileTimestamp())
        self._server.shutdown()
        self._profiler.record_event(shutdown_event)

    def __instance_id(self) -> Reference:
        """Returns our complete instance id.
        """
        return self._kernel + self._index

    def __get_client(self, instance: Reference) -> MPPClient:
        """Get or create a client to connect to the given instance.

        Args:
            instance: A reference to the instance to connect to.

        Returns:
            An existing or new MCP client.
        """
        if instance not in self._clients:
            locations = self._peer_info.get_peer_locations(instance)
            _logger.info(f'Connecting to peer {instance} at {locations}')
            self._clients[instance] = MPPClient(locations)

        return self._clients[instance]

    def __get_endpoint(self, port_name: str, slot: List[int]) -> Endpoint:
        """Determines the endpoint on our side.

        Args:
            port_name: Name of the port to send or receive on.
            slot: Slot to send or receive on.
        """
        try:
            port = Identifier(port_name)
        except ValueError as e:
            raise ValueError('"{}" is not a valid port name: {}'.format(
                port_name, e))

        return Endpoint(self._kernel, self._index, port, slot)

    def _close_port(self, port_name: str, slot: Optional[int] = None) -> None:
        """Closes the given port.

        This signals to any connected instance that no more messages
        will be sent on this port, which it can use to decide whether
        to shut down or continue running.

        Args:
            port_name: The name of the port to close.
        """
        message = Message(float('inf'), None, ClosePort(), Settings())
        if slot is None:
            _logger.debug('Closing port {}'.format(port_name))
        else:
            _logger.debug('Closing port {}[{}]'.format(port_name, slot))
        self.send_message(port_name, message, slot)

    def _close_outgoing_ports(self) -> None:
        """Closes outgoing ports.

        This sends a close port message on all slots of all outgoing
        ports.
        """
        for operator, ports in self._port_manager.list_ports().items():
            if operator.allows_sending():
                for port_name in ports:
                    port = self._port_manager.get_port(port_name)
                    if port.is_vector():
                        for slot in range(port.get_length()):
                            self._close_port(port_name, slot)
                    else:
                        self._close_port(port_name)

    def _drain_incoming_port(self, port_name: str) -> None:
        """Receives messages until a ClosePort is received.

        Receives at least once.

        Args:
            port_name: Port to drain.
        """
        port = self._port_manager.get_port(port_name)
        while port.is_open():
            # TODO: log warning if not a ClosePort
            self.receive_message(port_name)

    def _drain_incoming_vector_port(self, port_name: str) -> None:
        """Receives messages until a ClosePort is received.

        Works with (resizable) vector ports.

        Args:
            port_name: Port to drain.
        """
        port = self._port_manager.get_port(port_name)
        while not all([not port.is_open(slot)
                       for slot in range(port.get_length())]):
            for slot in range(port.get_length()):
                if port.is_open(slot):
                    self.receive_message(port_name, slot)

    def _close_incoming_ports(self) -> None:
        """Closes incoming ports.

        This receives on all incoming ports until a ClosePort is
        received on them, signaling that there will be no more
        messages, and allowing the sending instance to shut down
        cleanly.
        """
        for operator, port_names in self._port_manager.list_ports().items():
            if operator.allows_receiving():
                for port_name in port_names:
                    port = self._port_manager.get_port(port_name)
                    if not port.is_connected():
                        continue
                    if not port.is_vector():
                        self._drain_incoming_port(port_name)
                    else:
                        self._drain_incoming_vector_port(port_name)

    def _close_ports(self) -> None:
        """Closes all ports.

        This sends a close port message on all slots of all outgoing
        ports, then receives one on all incoming ports.
        """
        self._close_outgoing_ports()
        self._close_incoming_ports()



from typing import List, Optional, Tuple

import msgpack
from ymmsl import Reference

from libmuscle.mcp.protocol import RequestType
from libmuscle.mcp.transport_client import ProfileData, TransportClient, TimeoutHandler
from libmuscle.mcp.type_registry import transport_client_types


class MPPClient:
    """A client that connects to an MPP server.

    This client connects to a peer to retrieve messages. It uses an MCP
    Transport to connect.
    """
    def __init__(self, locations: List[str]) -> None:
        """Create an MPPClient for the given peer.

        The client will connect to the peer on one of its locations. It
        tries the most efficient protocol first. Once connected, it can
        request messages from any component and port represented by it.

        Args:
            locations: The peer's location strings
        """
        client: Optional[TransportClient] = None
        for ClientType in transport_client_types:
            for location in locations:
                if ClientType.can_connect_to(location):
                    try:
                        client = ClientType(location)
                        break
                    except Exception:
                        pass
            if client:
                break
        else:
            raise RuntimeError('Failed to connect')

        self._transport_client = client

    def receive(self, receiver: Reference, timeout_handler: Optional[TimeoutHandler]
                ) -> Tuple[bytes, ProfileData]:
        """Receive a message from a port this client connects to.

        Args:
            receiver: The receiving (local) port.

        Returns:
            The received message, and profiling data
        """
        request = [RequestType.GET_NEXT_MESSAGE.value, str(receiver)]
        encoded_request = msgpack.packb(request, use_bin_type=True)
        return self._transport_client.call(encoded_request, timeout_handler)

    def close(self) -> None:
        """Closes this client.

        This closes any connections this client has and/or performs
        other shutdown activities.
        """
        self._transport_client.close()



from typing import Callable, Set

import pytest

from libmuscle.api_guard import APIGuard


def test_no_checkpointing_support():
    guard = APIGuard(False)
    for _ in range(3):
        guard.verify_reuse_instance()
        guard.reuse_instance_done(True)

    guard.verify_reuse_instance()
    guard.reuse_instance_done(False)


def test_final_snapshot_only(guard: APIGuard):
    for i in range(4):
        guard.verify_reuse_instance()
        guard.reuse_instance_done(True)

        guard.verify_resuming()
        if i == 0:
            guard.resuming_done(True)

            guard.verify_load_snapshot()
            guard.load_snapshot_done()
        else:
            guard.resuming_done(False)

        guard.verify_should_init()
        guard.should_init_done()

        guard.verify_should_save_final_snapshot()
        if i == 2:
            guard.should_save_final_snapshot_done(True)

            guard.verify_save_final_snapshot()
            guard.save_final_snapshot_done()
        else:
            guard.should_save_final_snapshot_done(False)

    guard.verify_reuse_instance()
    guard.reuse_instance_done(False)


def test_full_checkpointing(guard: APIGuard):
    for i in range(4):
        guard.verify_reuse_instance()
        guard.reuse_instance_done(True)

        guard.verify_resuming()
        if i == 0:
            guard.resuming_done(True)

            guard.verify_load_snapshot()
            guard.load_snapshot_done()
        else:
            guard.resuming_done(False)

        guard.verify_should_init()
        guard.should_init_done()

        for j in range(3):
            guard.verify_should_save_snapshot()
            if j != 2:
                guard.should_save_snapshot_done(True)

                guard.verify_save_snapshot()
                guard.save_snapshot_done()
            else:
                guard.should_save_snapshot_done(False)

        guard.verify_should_save_final_snapshot()
        if i == 2:
            guard.should_save_final_snapshot_done(True)

            guard.verify_save_final_snapshot()
            guard.save_final_snapshot_done()
        else:
            guard.should_save_final_snapshot_done(False)

    guard.verify_reuse_instance()
    guard.reuse_instance_done(False)


_api_guard_funs = (
    (APIGuard.verify_reuse_instance, ()),
    (APIGuard.reuse_instance_done, (True,)),
    (APIGuard.verify_resuming, ()),
    (APIGuard.resuming_done, (True,)),
    (APIGuard.verify_load_snapshot, ()),
    (APIGuard.load_snapshot_done, ()),
    (APIGuard.verify_should_init, ()),
    (APIGuard.should_init_done, ()),
    (APIGuard.verify_should_save_snapshot, ()),
    (APIGuard.should_save_snapshot_done, (True,)),
    (APIGuard.verify_save_snapshot, ()),
    (APIGuard.save_snapshot_done, ()),
    (APIGuard.verify_should_save_final_snapshot, ()),
    (APIGuard.should_save_final_snapshot_done, (True,)),
    (APIGuard.verify_save_final_snapshot, ())
)


def run_until_before(guard: APIGuard, excluded: Callable) -> None:
    for fun, args in _api_guard_funs:
        if fun is excluded:
            break
        fun(guard, *args)


def check_all_raise_except(guard: APIGuard, excluded: Set[Callable]) -> None:
    for fun, args in _api_guard_funs:
        if fun.__name__.startswith('verify_'):
            if fun not in excluded:
                with pytest.raises(RuntimeError):
                    fun(guard, *args)
            else:
                fun(guard, *args)


@pytest.mark.parametrize('fun', [
        APIGuard.verify_load_snapshot,
        APIGuard.verify_should_init, APIGuard.verify_save_snapshot,
        APIGuard.verify_save_final_snapshot])
def test_missing_step(guard, fun):
    run_until_before(guard, fun)
    check_all_raise_except(guard, {fun})


def test_missing_resuming(guard: APIGuard):
    run_until_before(guard, APIGuard.verify_resuming)
    check_all_raise_except(guard, {APIGuard.verify_resuming})


def test_missing_should_save_final(guard: APIGuard):
    run_until_before(guard, APIGuard.verify_should_save_final_snapshot)
    check_all_raise_except(guard, {
        APIGuard.verify_should_save_snapshot,
        APIGuard.verify_should_save_final_snapshot})


def test_double_should_save(guard: APIGuard):
    run_until_before(guard, APIGuard.verify_should_save_snapshot)
    guard.verify_should_save_snapshot()
    guard.should_save_snapshot_done(True)
    with pytest.raises(RuntimeError):
        guard.verify_should_save_snapshot()



import cerulean
import logging
import pytest

from integration_test.cluster_test.conftest import (
        REMOTE_SHARED, ssh_term, skip_unless_cluster)


_logger = logging.getLogger(__name__)


@pytest.fixture
def fake_cluster(installed_cluster):
    headnode_port = installed_cluster
    term = ssh_term(headnode_port, 'Connection to virtual cluster container timed out')
    with cerulean.SftpFileSystem(term, True) as fs:
        local_sched = cerulean.DirectGnuScheduler(term)
        slurm_sched = cerulean.SlurmScheduler(term)
        yield term, fs, local_sched, slurm_sched


@pytest.fixture
def remote_home(fake_cluster):
    remote_fs = fake_cluster[1]
    return remote_fs / REMOTE_SHARED


@pytest.fixture
def remote_test_files(remote_home):
    return remote_home / 'cluster_test'


@pytest.fixture
def remote_out_dir(remote_home):
    return remote_home / 'test_results'


def _make_base_job(name, remote_out_dir, dir_name):
    job_dir = remote_out_dir / dir_name
    job_dir.mkdir(0o755, True, True)

    job = cerulean.JobDescription()
    job.name = name
    job.working_directory = job_dir
    job.stdout_file = job_dir / 'stdout.txt'
    job.stderr_file = job_dir / 'stderr.txt'
    job.queue_name = 'debug'
    job.time_reserved = 60
    job.system_out_file = job_dir / 'sysout.txt'
    job.system_err_file = job_dir / 'syserr.txt'
    job.extra_scheduler_options = '--ntasks-per-node=4'

    return job


def _make_job(name, mode, remote_test_files, remote_out_dir):
    job = _make_base_job(name, remote_out_dir, f'test_{name}_{mode}')
    job.command = str(remote_test_files / f'{name}.sh')
    return job


def _make_mpi_job(name, mode, execution_model, remote_test_files, remote_out_dir):
    job = _make_base_job(name, remote_out_dir, f'test_{name}_{mode}_{execution_model}')
    job.command = str(remote_test_files / f'{name}_{execution_model}.sh')
    return job


def _sched(fake_cluster, mode):
    if mode == 'local':
        return fake_cluster[2]
    else:
        return fake_cluster[3]


def _run_cmd_dir(remote_out_dir, testname, mode, execution_model=None):
    results_name = f'test_{testname}_{mode}'
    if execution_model is not None:
        results_name += f'_{execution_model}'

    for p in (remote_out_dir / results_name).iterdir():
        if p.name.startswith('run_'):
            return p


def _get_stdout(remote_out_dir, testname, mode, instance):
    run_dir = _run_cmd_dir(remote_out_dir, testname, mode)
    stdout_file = run_dir / 'instances' / instance / 'stdout.txt'
    assert stdout_file.exists()     # test output redirection
    return stdout_file.read_text()


def _get_outfile(remote_out_dir, testname, mode, execution_model, instance, rank):
    run_dir = _run_cmd_dir(remote_out_dir, testname, mode, execution_model)
    work_dir = run_dir / 'instances' / instance / 'workdir'
    out_file = work_dir / f'out_{rank}.txt'
    assert out_file.exists()        # test working directory
    return out_file.read_text()


_SCHED_OVERHEAD = 60


@skip_unless_cluster
@pytest.mark.parametrize('mode', ['local', 'slurm'])
def test_single(
        fake_cluster, remote_test_files, remote_out_dir, mode, hwthread_to_core):
    sched = _sched(fake_cluster, mode)

    job = _make_job('single', mode, remote_test_files, remote_out_dir)
    if mode == 'slurm':
        job.num_nodes = 1
        job.mpi_processes_per_node = 1
        job.extra_scheduler_options += ' --nodelist=node-0'

    job_id = sched.submit(job)
    assert sched.wait(job_id, job.time_reserved + _SCHED_OVERHEAD) is not None
    assert sched.get_exit_code(job_id) == 0

    output = _get_stdout(remote_out_dir, 'single', mode, 'c1')

    if mode == 'local':
        assert output.split('\n')[0] == 'headnode'
    else:
        node, hwthreads, _ = output.split('\n')
        assert node == 'node-0.example.org'
        assert hwthread_to_core(hwthreads) == [0]


@skip_unless_cluster
@pytest.mark.parametrize('mode', ['local', 'slurm'])
def test_dispatch(
        fake_cluster, remote_test_files, remote_out_dir, mode, hwthread_to_core):
    sched = _sched(fake_cluster, mode)

    job = _make_job('dispatch', mode, remote_test_files, remote_out_dir)
    if mode == 'slurm':
        job.num_nodes = 1
        job.mpi_processes_per_node = 1
        job.extra_scheduler_options += ' --nodelist=node-1'

    job_id = sched.submit(job)
    assert sched.wait(job_id, job.time_reserved + _SCHED_OVERHEAD) is not None
    assert sched.get_exit_code(job_id) == 0

    c1_out = _get_stdout(remote_out_dir, 'dispatch', mode, 'c1')
    c2_out = _get_stdout(remote_out_dir, 'dispatch', mode, 'c2')
    if mode == 'local':
        assert c1_out.split('\n')[0] == 'headnode'
        assert c2_out.split('\n')[0] == 'headnode'
    else:
        node, hwthreads, _ = c1_out.split('\n')
        assert node == 'node-1.example.org'
        assert hwthread_to_core(hwthreads) == [0]

        node, hwthreads, _ = c2_out.split('\n')
        assert node == 'node-1.example.org'
        assert hwthread_to_core(hwthreads) == [0]


@skip_unless_cluster
@pytest.mark.parametrize('mode', ['local', 'slurm'])
def test_multiple(
        fake_cluster, remote_test_files, remote_out_dir, mode, hwthread_to_core):
    sched = _sched(fake_cluster, mode)

    job = _make_job('multiple', mode, remote_test_files, remote_out_dir)
    if mode == 'slurm':
        job.num_nodes = 3
        job.extra_scheduler_options += ' --nodelist=node-[0-2]'

    job_id = sched.submit(job)
    assert sched.wait(job_id, job.time_reserved + _SCHED_OVERHEAD) is not None
    assert sched.get_exit_code(job_id) == 0

    for i in range(1, 7):
        instance = f'c{i}'
        out = _get_stdout(remote_out_dir, 'multiple', mode, instance)
        if mode == 'local':
            assert out.split('\n')[0] == 'headnode'
        else:
            node, hwthreads, _ = out.split('\n')
            assert (instance, node) == (instance, f'node-{(i - 1) // 2}.example.org')
            assert (instance, hwthread_to_core(hwthreads)) == (instance, [(i - 1) % 2])


@skip_unless_cluster
@pytest.mark.parametrize('mode', ['local', 'slurm'])
@pytest.mark.parametrize('execution_model', ['openmpi', 'intelmpi', 'srunmpi'])
def test_double(
        fake_cluster, remote_test_files, remote_out_dir, hwthread_to_core,
        mode, execution_model):

    if mode == 'local' and execution_model == 'srunmpi':
        pytest.skip('srun does not work without slurm')

    sched = _sched(fake_cluster, mode)

    job = _make_mpi_job(
            'double', mode, execution_model, remote_test_files, remote_out_dir)
    if mode == 'slurm':
        job.num_nodes = 2
        job.extra_scheduler_options += ' --nodelist=node-[3-4]'

    job_id = sched.submit(job)
    assert sched.wait(job_id, job.time_reserved + _SCHED_OVERHEAD) is not None
    assert sched.get_exit_code(job_id) == 0

    for i in range(1, 3):
        for rank in range(2):
            out = _get_outfile(
                    remote_out_dir, 'double', mode, execution_model, f'c{i}', rank)
            if mode == 'local':
                assert out.split('\n')[0] == 'headnode'
            else:
                node, hwthreads, _ = out.split('\n')
                assert node == f'node-{i + 2}.example.org'
                assert hwthread_to_core(hwthreads) == [rank]


@skip_unless_cluster
@pytest.mark.parametrize('mode', ['local', 'slurm'])
@pytest.mark.parametrize('execution_model', ['openmpi', 'intelmpi', 'srunmpi'])
def test_macro_micro(
        fake_cluster, remote_test_files, remote_out_dir, hwthread_to_core,
        mode, execution_model):

    if mode == 'local' and execution_model == 'srunmpi':
        pytest.skip('srun does not work without slurm')

    sched = _sched(fake_cluster, mode)

    job = _make_mpi_job(
            'macro_micro', mode, execution_model, remote_test_files, remote_out_dir)
    if mode == 'slurm':
        job.num_nodes = 1
        job.extra_scheduler_options += ' --nodelist=node-4'

    job_id = sched.submit(job)
    assert sched.wait(job_id, job.time_reserved + _SCHED_OVERHEAD) is not None
    assert sched.get_exit_code(job_id) == 0

    for i in range(1, 3):
        for rank in range(2):
            out = _get_outfile(
                    remote_out_dir, 'macro_micro', mode, execution_model, f'c{i}', rank)
            if mode == 'local':
                assert out.split('\n')[0] == 'headnode'
            else:
                node, hwthreads, _ = out.split('\n')
                assert node == 'node-4.example.org'
                assert hwthread_to_core(hwthreads) == [rank]


@skip_unless_cluster
@pytest.mark.parametrize('mode', ['local', 'slurm'])
def test_base_env(
        fake_cluster, remote_test_files, remote_out_dir, mode):
    sched = _sched(fake_cluster, mode)

    job = _make_job('base_env', mode, remote_test_files, remote_out_dir)
    if mode == 'slurm':
        job.num_nodes = 2
        job.mpi_processes_per_node = 2

    job_id = sched.submit(job)
    assert sched.wait(job_id, job.time_reserved + _SCHED_OVERHEAD) is not None
    assert sched.get_exit_code(job_id) == 0

    login_out = _get_stdout(remote_out_dir, 'base_env', mode, 'login')
    clean_out = _get_stdout(remote_out_dir, 'base_env', mode, 'clean')
    manager_out = _get_stdout(remote_out_dir, 'base_env', mode, 'manager')

    assert login_out.split('\n')[0] == '1'     # .profile loaded
    assert login_out.split('\n')[1] == '0'     # .bashrc loaded
    assert login_out.split('\n')[2] == '0'     # manager shell
    assert login_out.split('\n')[3] == '1'     # modules loaded
    assert login_out.split('\n')[4] == ''      # venv active

    assert clean_out.split('\n')[0] == '0'     # .profile loaded
    assert clean_out.split('\n')[1] == '0'     # .bashrc loaded
    assert clean_out.split('\n')[2] == '1'     # manager shell
    assert clean_out.split('\n')[3] == '0'     # modules loaded
    assert clean_out.split('\n')[4] == ''      # venv active

    assert manager_out.split('\n')[0] == '0'   # .profile loaded
    assert manager_out.split('\n')[1] == '0'   # .bashrc loaded
    assert manager_out.split('\n')[2] == '1'   # manager shell
    assert manager_out.split('\n')[3] == '1'   # modules loaded
    assert manager_out.split('\n')[4] == '/home/cerulean/shared/venv'



"""Module for describing compute resources

There's a huge comment here because there's a big mess here that took me forever to
figure out, so now I'm going to document it for the future.


Identifying hardware resources

Today's computers all contain multi-core CPUs, often with symmetric multithreading
(SMT), also known as hyperthreading. This means that we have hardware threads
(hwthreads) and also cores, and then there's caches and memory as well but we're not
going into NUMA here.

Cores and hwthreads are identified by number, but they have multiple different numbers
that are referred to by different names in different contexts, making everything very
confusing. So here are some definitions to disambiguate things.  Note that this is still
a rather simplified representation, but it's enough for what we're doing here in
MUSCLE3.


Hardware threads

A *hardware thread (hwthread)* is, at the hardware level, an instruction decoder. It
points to wherever in the code we are currently executing, and it can read the next
couple of instructions and figure out how to execute them. It can't actually execute
anything however, because it doesn't have the hardware that does that.

Intel refers to hwthreads as "logical processors" and so does Linux, hwloc calls them
"processing units" or PUs and so does OpenMPI unless it uses the term hwthread just to
confuse things a bit more.

Cores

A *core* contains at least one hwthread, and at least one functional unit, which is a
hardware component that actually does calculations and other data processing. Within a
core, the hwthread(s) read instructions and pass them to the functional units to be
executed. If a core has more than one hwthread, then the CPU supports SMT.

Intel refers to cores as "physical processors", hwloc calls them cores and so do most
other sources. We'll use cores here.

Since a hwthread cannot do anything on its own, it's always part of a core.

CPUs

The term CPU is used in many ways by various bits of documentation, sometimes referring
to a hwthread or a core, but here we'll take it to mean a collection of cores in a
plastic box. Similar terms are *package* (referring to that plastic box with very many
metal pins) and *socket* (the thing the package mounts into), or *processor*, which was
originally used to refer to all of the above when CPUs still had only one core with only
one hwthread, and has now become ambiguous.

Weird things can happen here, I've seen CPUs that as far as I can tell are a single
package, but nevertheless claim to have two sockets. I suspect that that's two physical
chips in a single plastic box, but I don't know for sure.

Here, we're concerned with hwthreads and cores and how to identify them and assign
instances to them.


Linux

On modern operating systems, hardware access is mediated by the operating system, and
we're mainly concerned with Linux here because that is what all the clusters are running
(see the note on macOS below). Information about the CPU(s) can be obtained on Linux
from the /proc/cpuinfo file, or equivalently but more modernly, from the files in
/sys/devices/system/cpu/cpu<x>/topology/.

Linux collects information about processors because it needs to run processes (programs,
software threads) on them on behalf of the user. Processes are assigned to hwthreads, so
that is what Linux considers a *processor*.  /proc/cpuinfo lists all these processors,
and they each have their own directory /sys/devices/system/cpu/cpu<x>.

On Linux, processors have an id, which is that number <x> in the directory, and is
listed under "processor" in /proc/cpuinfo. Since this number identifies a hwthread and
is assigned by Linux rather than being baked into the hardware, I'm calling it a
"logical hwthread id", this being a logical id of a hwthread, not an id of a logical
hwthread. It's also the id of a logical processor in Intel-speak.

Hwthreads actually have a second number associated with them, which does come from the
hardware. In /proc/cpuinfo, that's listed under "apicid"; it doesn't seem to be
available from sysfs. Hwloc call this the "physical PU (its name for a hwthread) id",
and OpenMPI's mpirun manpage also refers to it as a "physical processor location".

There's great potential for confusion here: the "physical PU id" and "physical processor
location" both identify a hardware-specified number (a physical id or a physical
location) for a hwthread. This is something completely different than what Intel calls a
"physical processor", which they use to refer to a core.

MUSCLE3 uses logical hwthread ids everywhere, it does not use physical ids.

Linux knows about how hwthreads are grouped into bigger things of course. Cores are
identified in Linux using the "core id", which is listed in /proc/cpuinfo and in
/sys/devices/system/cpu/cpu<x>/topology/core_id. So for each hwthread, identified by its
logical id, we can look up which core it is a part of. The core id is a logical id,
assigned by Linux, not by the hardware.  While logical hwthread ids seem to always be
consecutive at least on the hardware I've seen so far, core ids may have gaps.

MUSCLE3 does not use core ids, although it uses groups of hwthread ids that contain all
the hwthreads for a given core.


Resource binding

Running processes need something to run on, a hwthread. The assignment of process to
hwthread is done by the operating system's scheduler: when a process is ready to run,
the scheduler will try to find it a free hwthread to run on.

The scheduler can be constrained in which hwthreads it considers for a given process,
which is known as binding the process. This may have performance benefits, because
moving a process from one hwthread to another takes time. In MUSCLE3, when running on a
cluster, each process is assigned its own specific set of hwthreads to run on, and we
try to bind the instance to the assigned hwthreads.

Taskset

How this is done depends on how the instance is started. For non-MPI instances, we use a
Linux utility named 'taskset' that starts another program with a giving binding. The
binding is expressed as an *affinity mask*, a string of bits that say whether a given
processor (hwthread) can be used by the process or not. Each position in the string of
bits corresponds to the hwthread with that logical id.

OpenMPI

OpenMPI can bind cores in various ways, we use a rankfile and the --use-hwthread-cpus
option to specify the logical hwthread ids we want to bind each MPI process (rank) to.
Note that OpenMPI by default binds to cores, and can also bind to various other things
including sockets.

MPICH

MPICH doesn't support binding, as far as I can see.

Intel MPI

Intel MPI uses logical hwthread ids-based masks, specified in an environment variable,
to go with a machinefile that lists the nodes to put each process on.

Slurm srun

Slurm's srun has a CPU_BIND environment variable that likewise contains logical hwthread
ids-based masks, and a hostfile that lists the nodes to put each process on.

Here are some disambiguation tables to help with the confusion:


```
MUSCLE3     hwthread        logical hwthread id         physical hwthread id

Linux       processor       processor                   apicid
                                                        (/proc/cpuinfo only)

cgroups                     always uses these

taskset                     always uses these

hwloc       PU              PU L#<x>                    PU P#<x>

OpenMPI     hwthread        used in rankfile if         used in rankfile if
                            --use-hwthread-cpus         rmaps_rank_file_physical
                            is specified                MCA param set

Intel       logical         logical processor
            processor       number

srun                        used by --bind-to

psutil      logical         returned by Process.cpu_affinity()
            core            counted by psutil.cpu_count(logical=True)
```


```
MUSCLE3     core            core id

Linux       core            core id

Hwloc       core            core L#<x>

OpenMPI     core            used in rankfile if
                            --use-hwthread-cpus not
                            specified

psutil      physical        counted by psutil.cpu_count(logical=False)
            core
```

"""
from copy import copy, deepcopy
from typing import Dict, Iterable, Iterator, List, Optional, Set, Tuple


class Core:
    """Describes a CPU core or designates a core or one or more hwthreads.

    A core is a group of functional units with one or more instruction decoders. If the
    core supports symmetric multithreading (SMT, aka hyperthreading) then there will be
    more than one instruction decoder or hardware thread in the core.

    Note that the term "logical CPU" refers to an instruction decoder/hwthread. If the
    processor does not support SMT, then each core has a single decoder and so a logical
    CPU is also a core.

    This class can be used in different ways with slighly different interpretations.
    When describing hardware resources, it describes a core and all of its hwthreads. In
    this case, cid is the core id, and hwthreads contains the hwthread ids of all
    hwthreads on this core. If no SMT is supported, then there will be only one
    hwthread id.

    When designating a whole core (e.g. for use by a process), cid is set to the id of
    the core, and hwthreads contains all of the hwthreads on that core. When designating
    a hwthread on a particular core, cid is set to the id of the core and hwthreads
    contains the designated (single) hwthread.

    MUSCLE3 never assigns swthreads to subsets of hwthreads on a core, it assigns them
    to either a single hwthread or a single whole core. So if more than one hwthread is
    given, then we can assume that those are all the hwthreads on that core.

    Objects of this class automatically deepcopy when copied. This means that you can
    make a copy using copy.copy() and modify that copy anywhere without changing the
    original.

    Args:
        cid: ID of this core, to be used to refer to it
        hwthreads: Ids of hwthreads (logical CPUs) belonging to this core
    """
    def __init__(self, cid: int, hwthreads: Set[int]) -> None:
        """Create a Core"""
        self.cid = cid
        self.hwthreads = copy(hwthreads)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Core):
            return NotImplemented

        return self.cid == other.cid and self.hwthreads == other.hwthreads

    def __len__(self) -> int:
        return len(self.hwthreads)

    def __copy__(self) -> 'Core':
        return Core(self.cid, self.hwthreads)

    def __or__(self, other: object) -> 'Core':
        if not isinstance(other, Core):
            return NotImplemented

        if other.cid != self.cid:
            raise ValueError('Cannot merge hwthreads on different cores')

        return Core(self.cid, self.hwthreads | other.hwthreads)

    def __ior__(self, other: object) -> 'Core':
        if not isinstance(other, Core):
            return NotImplemented

        if other.cid != self.cid:
            raise ValueError('Cannot merge hwthreads on different cores')

        self.hwthreads |= other.hwthreads
        return self

    def __isub__(self, other: object) -> 'Core':
        if not isinstance(other, Core):
            return NotImplemented

        if other.cid != self.cid:
            raise ValueError('Cannot merge hwthreads on different cores')

        self.hwthreads -= other.hwthreads
        return self

    def __str__(self) -> str:
        hwthreads = ','.join(map(str, sorted(self.hwthreads)))
        return f'{self.cid}({hwthreads})'

    def __repr__(self) -> str:
        hwthreads = ','.join(map(str, sorted(self.hwthreads)))
        return f'Core({self.cid}, {{{hwthreads}}})'

    def isdisjoint(self, other: 'Core') -> bool:
        """Returns whether we share resources with other."""
        if self.cid != other.cid:
            raise ValueError('Cannot compare hwthreads on different cores')

        return self.hwthreads.isdisjoint(other.hwthreads)


class CoreSet:
    """A set of cores on a single node.

    This exists to make it a bit easier to operate on sets of cores, merging and
    subtracting them.

    Objects of this class automatically deepcopy when copied. This means that you can
    make a copy using copy.copy() and modify that copy anywhere without changing the
    original.
    """
    def __init__(self, cores: Iterable[Core]) -> None:
        """Create a CoreSet

        Args:
            cores: A set of cores to contain.
        """
        self._cores = {c.cid: c for c in cores}

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, CoreSet):
            return NotImplemented

        if len(self._cores) != len(other._cores):
            return False

        for cid, core in self._cores.items():
            if cid not in other._cores:
                return False
            if core.hwthreads != other._cores[cid].hwthreads:
                return False

        return True

    def __len__(self) -> int:
        return len(self._cores)

    def __iter__(self) -> Iterator[Core]:
        return iter(self._cores.values())

    def __copy__(self) -> 'CoreSet':
        return CoreSet(deepcopy(list(self._cores.values())))

    def __ior__(self, other: object) -> 'CoreSet':
        if not isinstance(other, CoreSet):
            return NotImplemented

        for cid, core in other._cores.items():
            if cid in self._cores:
                self._cores[cid] |= core
            else:
                self._cores[cid] = copy(core)

        return self

    def __isub__(self, other: object) -> 'CoreSet':
        if not isinstance(other, CoreSet):
            return NotImplemented

        for cid, core in other._cores.items():
            if cid in self._cores:
                self._cores[cid] -= core
                if not self._cores[cid].hwthreads:
                    del self._cores[cid]

        return self

    def __str__(self) -> str:
        def collapse_ranges(ids: List[int]) -> str:
            if len(ids) == 0:
                return ''

            result = list()
            start = 0
            i = 1
            while i <= len(ids):
                if (i == len(ids)) or (ids[i-1] != ids[i] - 1):
                    if start == i - 1:
                        # run of one
                        result.append(str(ids[i-1]))
                    else:
                        # run of at least two
                        result.append(f'{ids[start]}-{ids[i-1]}')
                    start = i
                i += 1
            return ','.join(result)

        cores = sorted((c.cid for c in self._cores.values()))
        hwthreads = sorted((t for c in self._cores.values() for t in c.hwthreads))

        return f'{collapse_ranges(cores)}({collapse_ranges(hwthreads)})'

    def __repr__(self) -> str:
        cores = ', '.join(map(repr, sorted(self._cores.values(), key=lambda c: c.cid)))
        return f'CoreSet({{{cores}}})'

    def isdisjoint(self, other: 'CoreSet') -> bool:
        """Returns whether we share resources with other."""
        for cid, core in self._cores.items():
            if cid in other._cores:
                if not core.isdisjoint(other._cores[cid]):
                    return False
        return True

    def get_first_cores(self, num_cores: int) -> 'CoreSet':
        """Returns the first num_cores cores in this set.

        Args:
            The number of cores to select.
        """
        result = copy(self)
        cids = list(self._cores.keys())
        selected = cids[:num_cores]
        if len(selected) < num_cores:
            raise RuntimeError('Tried to get more cores than available')

        result._cores = {c.cid: c for c in result._cores.values() if c.cid in selected}
        return result


class OnNodeResources:
    """Resources on a single node, currently only CPU cores.

    This represents a set of resources on a single node, either all of the resources
    available or some subset of interest.

    Objects of this class automatically deepcopy when copied. This means that you can
    make a copy using copy.copy() and modify that copy anywhere without changing the
    original.
    """
    def __init__(self, node_name: str, cpu_cores: CoreSet) -> None:
        """Create an OnNodeResources.

        Args:
            name: (Host)name of the node.
            cpu_cores: A set of cores for this node.
        """
        self.node_name = node_name
        self.cpu_cores = cpu_cores

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, OnNodeResources):
            return NotImplemented

        return (
                isinstance(other, OnNodeResources) and
                self.node_name == other.node_name and
                self.cpu_cores == other.cpu_cores)

    def __copy__(self) -> 'OnNodeResources':
        return OnNodeResources(self.node_name, copy(self.cpu_cores))

    def __ior__(self, other: object) -> 'OnNodeResources':
        if not isinstance(other, OnNodeResources):
            return NotImplemented

        if self.node_name != other.node_name:
            raise ValueError('Cannot merge resources on different nodes')

        self.cpu_cores |= other.cpu_cores
        return self

    def __isub__(self, other: object) -> 'OnNodeResources':
        if not isinstance(other, OnNodeResources):
            return NotImplemented

        if self.node_name != other.node_name:
            raise ValueError('Cannot remove resources on different nodes')

        self.cpu_cores -= other.cpu_cores
        return self

    def __str__(self) -> str:
        return f'OnNodeResources({self.node_name}, c: {str(self.cpu_cores)})'

    def __repr__(self) -> str:
        return f'OnNodeResources("{self.node_name}", {repr(self.cpu_cores)})'

    def hwthreads(self) -> Iterable[int]:
        """Return the hwthreads in this node."""
        return (thread for core in self.cpu_cores for thread in core.hwthreads)

    def total_cores(self) -> int:
        """Return the number of CPU cores in this node."""
        return len(self.cpu_cores)

    def isdisjoint(self, other: 'OnNodeResources') -> bool:
        """Returns whether we share resources with other."""
        return (
                self.node_name != other.node_name or
                self.cpu_cores.isdisjoint(other.cpu_cores))


class Resources:
    """Designates a (sub)set of resources.

    Whether these resources are free or allocated in general or by something specific
    depends on the context, this just says which resources we're talking about.

    Objects of this class automatically deepcopy when copied. This means that you can
    make a copy using copy.copy() and modify that copy anywhere without changing the
    original.

    Attributes:
        nodes: A collection of nodes to include in this resource set
    """
    def __init__(self, nodes: Optional[Iterable[OnNodeResources]] = None) -> None:
        """Create a Resources object with the given nodes.

        Args:
            nodes: OnNodeResourcess to be designated by this object.
        """
        if nodes is None:
            self._nodes: Dict[str, OnNodeResources] = {}
        else:
            self._nodes = {n.node_name: n for n in nodes}

    def __len__(self) -> int:
        return len(self._nodes)

    def __iter__(self) -> Iterator[OnNodeResources]:
        return iter(self._nodes.values())

    def __getitem__(self, node_name: str) -> OnNodeResources:
        return self._nodes[node_name]

    def __eq__(self, other: object) -> bool:
        """Check for equality."""
        if not isinstance(other, Resources):
            return NotImplemented

        if len(self._nodes) != len(other._nodes):
            return False

        for node_name, node in self._nodes.items():
            if node_name not in other._nodes:
                return False
            if other._nodes[node_name] != node:
                return False

        return True

    def __copy__(self) -> 'Resources':
        """Copy the object."""
        return Resources((copy(n) for n in self._nodes.values()))

    def __ior__(self, other: object) -> 'Resources':
        """Add the resources in the argument to this object."""
        if not isinstance(other, Resources):
            return NotImplemented

        for node_name, other_node in other._nodes.items():
            if node_name in self._nodes:
                self._nodes[node_name] |= other_node
            else:
                self._nodes[node_name] = copy(other_node)

        return self

    def __isub__(self, other: object) -> 'Resources':
        """Remove the resources in the argument from this object."""
        if not isinstance(other, Resources):
            return NotImplemented

        for node_name, other_node in other._nodes.items():
            if node_name in self._nodes:
                self._nodes[node_name] -= other_node
                if not self._nodes[node_name]:
                    del self._nodes[node_name]

        return self

    def __str__(self) -> str:
        """Return a human-readable string representation."""
        nodes = ','.join(
                map(str, sorted(self._nodes.values(), key=lambda n: n.node_name)))
        return f'Resources({nodes})'

    def __repr__(self) -> str:
        """Return a string representation."""
        nodes = sorted(self._nodes.values(), key=lambda n: n.node_name)
        return f'Resources({nodes})'

    def nodes(self) -> Iterable[str]:
        """Return the names of the nodes on which we designate resources."""
        return self._nodes.keys()

    def total_cores(self) -> int:
        """Return the total number of cores (not hwthreads) designated."""
        return sum((len(n.cpu_cores) for n in self._nodes.values()))

    def cores(self) -> Iterable[Tuple[str, int]]:
        """Return this resources as a list of node, core."""
        return (
                (node.node_name, core.cid)
                for node in self._nodes.values() for core in node.cpu_cores)

    def hwthreads(self) -> Iterable[Tuple[str, int]]:
        """Return this resources as a list of node, hwthread."""
        return (
                (node.node_name, hwthread)
                for node in self._nodes.values() for hwthread in node.hwthreads())

    def isdisjoint(self, other: 'Resources') -> bool:
        """Return whether we share resources with other."""
        for node_name, node in self._nodes.items():
            if node_name in other._nodes:
                if not node.isdisjoint(other._nodes[node_name]):
                    return False
        return True

    def add_node(self, node_res: OnNodeResources) -> None:
        """Add a node's resources.

        This absorbs node_res into this Resources object, so if you change node_res
        after adding it, the changes will be reflected in this Resources.

        Args:
            node_res: Resources on a node not yet included in this Resources.

        Raises:
            RuntimeError: if we already have a node with this node name.
        """
        if node_res.node_name in self._nodes:
            raise RuntimeError(
                    'Tried to add a OnNodeResources to a Resources for a node that is'
                    ' already present. This is a bug in MUSCLE3, please report it on'
                    ' GitHub.')

        self._nodes[node_res.node_name] = node_res

    def merge_node(self, node_res: OnNodeResources) -> None:
        """Merges a node's resources

        This always copies the object.

        Args:
            node_res: Resources on a node that may already be included in this
                    Resources.
        """
        if node_res.node_name in self._nodes:
            self._nodes[node_res.node_name] |= node_res
        else:
            self._nodes[node_res.node_name] = copy(node_res)

    @staticmethod
    def union(resources: Iterable['Resources']) -> 'Resources':
        """Combines the resources into one.

        Args:
            resources: A collection of resources to merge.

        Return:
            A Resources object referring to all the resources in the
            input.
        """
        result = Resources()
        for cur_resources in resources:
            result |= cur_resources
        return result
